# -*- coding: utf-8 -*-
# package: pypos3d.wftk
import sys, copy
import logging
import math
import array
import numpy as np
from scipy import spatial

from langutil import C_FAIL, C_ERROR , C_OK

from pypos3d.wftk import WFBasic
from pypos3d.wftk.WFBasic import C_VERTEX_NOT_FOUND, COUNT, FEPSILON, C_FACE_NORM, \
  C_FACE_TEXT, TexCoord2f, CoordSyst, C_FACE_MATMASK, C_FACE_TEXTNORM, FEPSILON2, C_FACE_SURF, \
  C_FACE_ORDER_ANGLE, C_FACE_ORDER_LENGTH, C_FACE_ORDER_SURF, C_FACE_NONE
from pypos3d.wftk.WFBasic import LowestIdxPos, CreateLoop, FaceNormalOrder
from pypos3d.wftk.WFBasic import Point3d, Vector3d, Edge, Regularity, IndexAdd, CreateTransform, MxProd
from pypos3d.wftk.PaveList2D import PaveList2D


class FaceAttr:
  ''' This class represent Face attributs.
  In a GeomGroup each face has its own FaceAttr to store:
  - fAttr (integer):
    - Material index (reached by C_FACE_MATMASK)
    - Flags : C_FACE_TEXT | C_FACE_NORM
  
  - tvertIdx : An array of 32bits integer to store texture indexes (if any else None)
  - normIdx : An array of 32bits integer to store normal indexes (if any else None)
  '''
  def __init__(self, fAttrOrSrc:'int or FaceAttr'=None, tvertIdx=None, normIdx=None):
    if isinstance(fAttrOrSrc, FaceAttr):
      self.fAttr = fAttrOrSrc.fAttr
      self.tvertIdx = copy.copy(fAttrOrSrc.tvertIdx) if fAttrOrSrc.tvertIdx else None
      self.normIdx = copy.copy(fAttrOrSrc.normIdx) if fAttrOrSrc.normIdx else None
    else:
      self.fAttr = fAttrOrSrc
      self.tvertIdx = tvertIdx if isinstance(tvertIdx, array.array) else (array.array('l', tvertIdx) if tvertIdx else None)
      self.normIdx = normIdx if isinstance(normIdx, array.array) else (array.array('l', normIdx) if normIdx else None)

  def __str__(self):
    return str(self.fAttr)

  def setMatIdx(self, idx):
    self.fAttr = (self.fAttr & C_FACE_TEXTNORM) + (idx & C_FACE_MATMASK)



class GeomGroup(object):
  ''' GeomGroup represents a Geometrical group of faces and lines (3D).  
  New implementation based on array.array on Long (32bits)
  '''

  def __init__(self, name='', src=None, inCoordList=None):
    ''' Create a GeomGroup.
    
    Parameters
    ----------
    name : str
      Name of the GeomGroup

    src : GeomGroup
      Source for the GeomGroup (copy constructor)

    inCoordList : list of Point3d
      List of 3D vertex to initialize the GeomGroup
    '''
    #  Pointer to the coordinate table of the whole object
    self.coordList = src.coordList if src else (inCoordList if inCoordList else [])
    self.texList   = src.texList   if src else [ ]
    self.normList  = src.normList  if src else [ ]

    # Pointer to the WaveGeom
    self.geom = src.geom if src else None

    # group name. Multiple group names are not managed as defined in OBJ format
    self._name = src._name if src else name

    # Contains the index of the first element of vertIdx that defines a face
    # Length > 1
    self.stripCount = copy.copy(src.stripCount) if src else array.array('l', [ 0, ])
    
    # Contains the index of the first element of vertIdx that defines a line
    # Null if no line OR Length > 1
    self.lineStripCount = copy.copy(src.lineStripCount) if src else array.array('l', [ 0, ])

    # Declare Faces attributs (C_FACE_NONE, C_FACE_TEXT, C_FACE_NORM, C_FACE_TEXTNORM, C_FACE_MATMSASK)
    # As of now, each face may have independently textures or normals 
    self.matIdx = copy.deepcopy(src.matIdx) if src else [ ]

    # Indexes of Vertex that compose each face
    self.vertIdx =  copy.copy(src.vertIdx) if src else array.array('l')

    # Indexes of Vertex that compose each line
    self.vertLineIdx =  copy.copy(src.vertLineIdx) if src else array.array('l')
    
    # Declare Lines attributs (C_FACE_NONE, C_FACE_TEXT, C_FACE_MATMSASK)
    # As of now, each line may have independently textures
    self.matLineIdx = copy.deepcopy(src.matLineIdx) if src else [ ]


    self._bestCrit = src._bestCrit if src else 0.0
    self.curMatIdx = src.curMatIdx if src else 0
    self.lstMat = src.lstMat if src else None

  def __str__(self):
    return f'GeomGroup[{self._name}]: {self.getNbFace()} faces'


  def getNbFace(self):
    ''' Returns the number of faces '''
    return len(self.stripCount) - 1

  def getNbLine(self):
    ''' Returns the number of lines '''
    return len(self.lineStripCount) - 1

  def getName(self):  return self._name

  def setName(self, n):
    self._name = n

  def getFaceStartIdx(self, noface):
    return self.stripCount[noface]

  def getFaceLastIdx(self, noface):
    return self.stripCount[noface + 1]

  def getMatIdx(self, noface):
    return self.matIdx[noface].fAttr & C_FACE_MATMASK

  def setMatIdx(self, noface, idx):
    self.matIdx[noface].setMatIdx(idx)

  # public int[] getFaceVertIdx(int noface, int[] restab)
  def getFaceVertIdx(self, noface):
    ''' Return the list of index of the Vertex of the given face. '''
    return self.vertIdx[self.stripCount[noface]:self.stripCount[noface + 1]] 

  # public int[] getFaceTVertIdx(int noface)
  def getFaceTVertIdx(self, noface):
    ''' Return the list of index of the texture coordinates of the given face.
    Return None if the face has no texture
    '''
    return self.matIdx[noface].tvertIdx

  def getFaceNormIdx(self, noface):
    ''' Return the list of index of the normals (Vector3d) of the given face.
    Return None if the face has no normal
    '''
    return self.matIdx[noface].normIdx 


  def getFaceVertex(self, noface, copyPoint=False, restab=None):
    ''' Return the list of Vertex of given face.
    Parameters
    ----------
    noface : int
      Face number [0 .. NbFace[
    copyPoint : bool
      if true, new Point3d are created.
      By default, this method return a pointers to the existing Point3d (in CoordList)
    restab : list
      Use restab to populate the result, else create a new list

    Returns
    -------
    list of Point3d
    '''
    startIdx = self.getFaceStartIdx(noface)
    lastIdx = self.getFaceLastIdx(noface)
    
    if restab:
      restab[:] = [ Point3d(self.coordList[j]) for j in self.vertIdx[startIdx:lastIdx] ] if copyPoint else [ self.coordList[j] for j in self.vertIdx[startIdx:lastIdx] ]
      return restab

    return [ Point3d(self.coordList[j]) for j in self.vertIdx[startIdx:lastIdx] ] if copyPoint else [ self.coordList[j] for j in self.vertIdx[startIdx:lastIdx] ]


  def getFaceVertexBy(self, matName, raiseExcept=True):
    ''' Return the list of Vertex of faces with the given material name.
    When raiseExcept is True:
      Raise ValueError when the material does not exist or the list is empty
    Else return an empty list
    
    Parameters
    ----------
    matName : str
      Material Name
      
    raiseExcept : bool, optional, default True
      if true, raise a ValueError
      else may return a None list

    Returns
    -------
    list of Point3d
    '''
    try:
      matidx = self.lstMat.index(matName)
      lstFaceNo = [ fi for fi in range(0, self.getNbFace()) if (self.matIdx[fi].fAttr&C_FACE_MATMASK)==matidx ]    
      
      if lstFaceNo:
        return [ self.coordList[j] for j in set(i for faceno in lstFaceNo for i in self.vertIdx[self.stripCount[faceno]:self.stripCount[faceno+1]]) ]
      
    except ValueError:
      pass

    if raiseExcept:
      raise ValueError('No face for '+matName)      
    else:
      return None

  #
  #
  def getFaceLoop(self, faceDescr, invertNorm=False):
    ''' Returns an (eventually) textured set of Edges
    Point3d are copied from the coordList and muted with a .texture attr if the face
    is textured

    Parameters
    ----------
    faceDescr : int or str
      if faceDescr is a str, it is considered as a Material Name. The selected face, is the first 
      with this material
      if faceDescr is an int,it should be a valid faceno
    invertNorm : bool
      if true the face order is reverted
    
    Returns
    -------
    list of Edge
    '''
    faceno = self.findFace(self.geom.lstMat.index(faceDescr)) if isinstance(faceDescr, str) else int(faceDescr)

    # Create the list of edges with new vertex with potential textures 
    startidx = self.getFaceStartIdx(faceno)
    lastidx = self.getFaceLastIdx(faceno)
    
    # Use new points with previous face definition
    vxtab = [ Point3d(self.coordList[self.vertIdx[i]]) for i in range(startidx, lastidx) ]

    if invertNorm:
      vxtab.reverse()

    fao = self.matIdx[faceno]
    hasTexture = fao.fAttr & C_FACE_TEXT != 0 
    if hasTexture: # Mutate the Point3d, so that they contain the associated texture coordinates
      for i, p in enumerate(vxtab):
        p.texture = self.texList[fao.tvertIdx[i]]
  
    # Return the list of face edges
    return [ Edge(vxtab[i], vxtab[i+1], -1, -1, hasTexture) for i in range(0, len(vxtab)-1) ] + [ Edge(vxtab[-1], vxtab[0]), ]



  def findTVertIdx(self, ptIdx):
    ''' Find the first face that refers to <code>vertIdx</code> and return the associated texture index.
    Parameters
    ----------
    vertIdx : int
      Index of the vertex to look for.

    Returns
    -------
    int
      The texture index, 
       C_VERTEX_NOT_FOUND if vertex not found, 
       C_FAIL if no texture index are available
    '''
    for faceno in range(0, len(self.stripCount) - 1):
      startIdx = self.stripCount[faceno]
      lastIdx = self.stripCount[faceno + 1]

      for i in range(startIdx, lastIdx):
        if self.matIdx[faceno].tvertIdx[i] == ptIdx:
          return ptIdx

    return C_VERTEX_NOT_FOUND

  def addFace(self, coordIdx, texIdx, normIdx):
    ''' Add a face defined by indexes of Vertex, Texture Coords and normals if any.
  
    Return the origin coordIdx for convenience
    '''   
    if not isinstance(coordIdx, array.array):
      coordIdx = array.array('l', coordIdx)

    if texIdx and not isinstance(texIdx, array.array):
      texIdx = array.array('l', texIdx)

    if normIdx and not isinstance(normIdx, array.array):
      normIdx = array.array('l', normIdx)
    
    self.vertIdx += coordIdx
    self.stripCount.append(len(self.vertIdx))
    fao = FaceAttr(self.curMatIdx | (C_FACE_TEXT if texIdx else 0) | (C_FACE_NORM if normIdx else 0), texIdx, normIdx)
    self.matIdx.append(fao)
        
    return coordIdx


  def addLine(self, coordIdx, texIdx): # List<Integer> coordIdx, List<Integer> texIdx)
    ''' Add a line defined by indexes of Vertex, Texture Coords and normals if any.
  
    Return the origin coordIdx for convenience
    '''   
    if not isinstance(coordIdx, array.array):
      coordIdx = array.array('l', coordIdx)

    if not isinstance(texIdx, array.array):
      texIdx = array.array('l', texIdx)

    self.vertLineIdx += coordIdx
    self.lineStripCount.append(len(self.vertLineIdx))
    lao = FaceAttr(self.curMatIdx | (C_FACE_TEXT if texIdx else 0))
    self.matLineIdx.append(lao)
    return coordIdx
    
  def addFaceByEdges(self, lstEdges, hasTexture, materialName, refNorm=None):
    ''' Create the closing face(s) and add it(them) to the list of faces of the group.
    This method uses the CreateLoop function based on a min/max tree search with 
    alpha-beta prune. 

    Parameters
    ----------
    lstEdges : list of Edge(s)
      List of Edges where faces will be search
    hasTexture : bool
      Indicates that the face has texture coordinates
    materialName : str
      Material to assign to the created face(s)
    refNorm : Vector3d
      if not null used to reorganize faces so that their normals are aligned with the reference one

    Returns
    -------
    list of face indexes, number of faces, list of Point3d
    '''
    wg = self.geom
    self.curMatIdx = wg.addMaterial(materialName)
    
    lstFacesIdx = [ ]
  
    FaceLst, coordList, *_ = CreateLoop([ e for e in lstEdges if e])
    logging.info('Loop %d faces created', len(FaceLst))
    # Warning: Face edges' order my be mixed
    for f in FaceLst:
      
      if refNorm and f: # Check face definition : Counter clock shall match refNorm
        FaceNormalOrder(f, refNorm)
      
      coordIdx = array.array('l', [ IndexAdd(wg.coordList, e.p0) for e in f ])
      lstFacesIdx += coordIdx
      texIdx = array.array('l', [ IndexAdd(wg.texList, e.p0.texture) for e in f ] if hasTexture else [ ] )
        
      self.addFace(coordIdx, texIdx, [ ])
  
    return lstFacesIdx, len(FaceLst), coordList


  def addFacesByVertexO2(self, nFaceList, nMatList):
    ''' Add a list of faces defined by a list of vertex.
    O(n2) algorithm
    nMatList : array of Long
    '''    
    for faceno, f in enumerate(nFaceList):
      coordIdx, texIdx  = [], [] # array.array('l'), array.array('l')
      
      # Protect from "None" faces with Slice Option
      if not f: continue
      
      for p in f:
        try:
          idx = self.coordList.index(p)
        except:
          idx = len(self.coordList)
          self.coordList.append(p)
          
        coordIdx.append(idx)
        
        if nMatList[faceno] & C_FACE_TEXT:
          try:
            idx = self.texList.index(p.texture)
          except:
            idx = len(self.texList)
            self.texList.append(p.texture)
            
          texIdx.append(idx)  
          
      # Take into account original Material of the face
      self.curMatIdx = nMatList[faceno] & C_FACE_MATMASK
      self.addFace(coordIdx, texIdx, [ ])
  
  def addFacesByVertex(self, nFaceList, nMatList):
    ''' Add a list of faces defined by a list of vertex.
    Octree (scipy.spatial.KDTree) algorithm
    nMatList : array of Long
    '''    
    # List of new points
    tmpCoordList = [ ]
    nbMaxVerts = len(self.coordList)
    nbMaxTVerts = len(self.texList)
    
    if (nbMaxVerts<1000) and (nbMaxTVerts<1000):
      logging.info("Using O(n2) method: %d", nbMaxVerts)
      return self.addFacesByVertexO2(nFaceList, nMatList)
    
    logging.info("Init CoordList/TexList: %d/%d", nbMaxVerts, nbMaxTVerts)
 
    nbsrc = len(self.coordList)
    npTab = np.zeros( (nbsrc, 3) )
    for pNo, p in enumerate(self.coordList):
      npTab[pNo] = [ p.x, p.y, p.z ]
    
    # Create an KDTree with the 'known Vertex' in a "global" np.array
    tree = spatial.KDTree(npTab, leafsize=10 if nbsrc<10000 else 100)
    svect = np.zeros((1,3))
    
    # Create a PaveList with the 'known Texture coords' in a "global" np.array
    pl = PaveList2D(n=32, texList=self.texList)
    
    for faceno, f in enumerate(nFaceList):
      coordIdx, texIdx = [], [] # array.array('l'), array.array('l')
      
      # Protect from "None" faces with Slice Option
      if not f: continue
      
      for p in f:
        # Search in KDTtree
        svect[0] = [ p.x, p.y, p.z ]
        rest, resIdx = tree.query(svect)

        # if found (not too far) ==> Put it in a tmp table
        if rest[0]<FEPSILON:
          # Use an existing vertex
          idx = resIdx[0]
        else:
          # Add a new vertex to the local list of new Vertex
          try:
            idx = tmpCoordList.index(p)
          except:
            idx = len(tmpCoordList)
            tmpCoordList.append(p)

          # New Points are comming after the existing ones
          idx+=nbsrc          

        coordIdx.append(idx)
        
        if nMatList[faceno] & C_FACE_TEXT:
          idx = pl.IndexAdd(p.texture)
          texIdx.append(idx)  
          
      # Take into account original Material of the face
      self.curMatIdx = nMatList[faceno] & C_FACE_MATMASK
      self.addFace(coordIdx, texIdx, [ ])
    #End for faceno 

    # Update the Global CoordList
    self.coordList += tmpCoordList



  # public int fillDedupFace(int nbFace, int[] tabHshFace, int[][] tabFaceIdx)
  def fillDedupFace(self, nbFace, tabHshFace, tabFaceIdx):
    ''' Copy strip data to deduplication tables '''
    for faceno in range(0, len(self.stripCount) -1):
      startIdx = self.stripCount[faceno]
      lastIdx = self.stripCount[faceno+1]
      idxTab = self.vertIdx[startIdx:lastIdx]
      nbv = lastIdx - startIdx

      # Find lowest index
      lowestIdx = LowestIdxPos(idxTab)
      finalIdxTab = [ idxTab[(i + lowestIdx) % nbv] for i in range(0, nbv) ]
      tabHshFace.append( sum(finalIdxTab) )
      tabFaceIdx.append( finalIdxTab )
      nbFace+=1

    return nbFace

  def extendLoopVicinity(self, edgePtIdx, edgeLoopPtIdx):
    '''Extends edgeLoopPtIdx with vertex of faces that are using vertex of edgePtIdx tab.
     As of 23-May-2020 : edgeLoopPtIdx does not need to be initialized. It'll grow 
     @param edgePtIdx       Initial table of vertex
     @param edgeLoopPtIdx   Table to extends.
    '''

    # For each point of existing loop
    for vi in edgePtIdx:
      # Search for faces that contains the current vertex
      for fno in range(0, self.getNbFace()):
        staIdx = self.stripCount[fno]
        lstIdx = self.stripCount[fno+1]

        if vi in self.vertIdx[staIdx:lstIdx]:
          # Add all vertex of the face to the target Loop if the new vertex is not already in it
          for nvertidx in self.vertIdx[staIdx:lstIdx]:
            if not nvertidx in edgeLoopPtIdx:
              edgeLoopPtIdx.append(nvertidx)


#  public long[] buildVectors()
  def buildVectors(self, geomcond=None):
    vectTbl = [ ]

    for faceno in range(0, len(self.stripCount) - 1):
      startIdx = self.stripCount[faceno]
      lastIdx = self.stripCount[faceno + 1] - 1
      
      fao = self.matIdx[faceno]
      hasTexture = fao.fAttr & C_FACE_TEXT
      
      for i in range(startIdx, lastIdx):
        idx0 = self.vertIdx[i]
        idx1 = self.vertIdx[i + 1]
        p0 = Point3d(self.coordList[idx0])
        p1 = Point3d(self.coordList[idx1])

        if not geomcond or (geomcond(p0) and geomcond(p1)):
          if hasTexture:
            p0.texture = self.texList[fao.tvertIdx[i-startIdx]]
            p1.texture = self.texList[fao.tvertIdx[i+1-startIdx]]
          vectTbl.append(Edge(p0, p1, idx0, idx1))

      # To close the loop
      idx0 = self.vertIdx[lastIdx]
      idx1 = self.vertIdx[startIdx]
      p0 = Point3d(self.coordList[idx0])
      p1 = Point3d(self.coordList[idx1])
      if not geomcond or (geomcond(p0) and geomcond(p1)):
        if hasTexture:
          p0.texture = self.texList[fao.tvertIdx[lastIdx-startIdx]]
          p1.texture = self.texList[fao.tvertIdx[0]]
        vectTbl.append(Edge(p0, p1, idx0, idx1))

    # Sort the Vector table to keep only lonely vectors
    return sorted(vectTbl, key=Edge.hashCode)


  def findEdges(self, geomcond=None):
    ''' Find a set of edges (according geomcond on Point3d)
    Edge are containg a copy of each source point (with a .texture attribut if any)
    '''
    logging.info("Group[%s]: %d faces", self._name , self.getNbFace())
    vectTbl = self.buildVectors(geomcond=geomcond) # Edge[] 
    logging.info("Group[%s]: %d vectors", self._name, len(vectTbl))

    vectno = 0

    # Edge[] edgeVect
    edgeVect = [ ]

    lenin = len(vectTbl)
    while vectno < lenin:
      curVect = vectTbl[vectno]
      vectno+=1

      nbocc = 1
      while (vectno < lenin) and (curVect==vectTbl[vectno]):
        vectno+=1
        nbocc+=1

      if nbocc==1:
        edgeVect.append(curVect)

    logging.info("Group[" + self._name + "]: " + str(len(edgeVect)) + " edge vectors deduced")

    return edgeVect

  def findFace(self, matidx):
    ''' Return the first faceno using the given material (by index) '''
    for i,fao in enumerate(self.matIdx):
      if fao.fAttr & C_FACE_MATMASK == matidx:
        return i
       
    return -1

  def setBestCrit(self, crit):
    self._bestCrit = crit

  def setMaterialName(self, matIdx):
    self.curMatIdx = matIdx

  def invertFaceOrder(self):
    ''' Invert all faces order (ie. revert 'auto' normals)
    
    Return
    ------
    self for convenience
    '''
    tmpVert = [] # array.array('l')

    for faceno in range( 0, self.getNbFace()):
      fvi = self.getFaceVertIdx(faceno)
      fvi.reverse()
      tmpVert += fvi
      
      fao = self.matIdx[faceno]
      if fao.fAttr & C_FACE_TEXT:
        fao.tvertIdx.reverse()
        
    self.vertIdx[:] = array.array('l', tmpVert)
      
    return self

    


                         
    
  # -----------------------------------------------------------------------------
  def sanityCheck(self):
    ''' Check the consistency of a group. '''
    res = C_OK
    name = self.getName() if self.getName() else "NoName"
    
    sizeVertexList = len(self.geom.coordList)
    sizeTexList = len(self.geom.texList)
    sizeNormList = len(self.geom.normList)
    
    # Verify Vertex indexes
    for vi in self.vertIdx:
      if (vi<0) or (vi>=sizeVertexList):
        print(f'Group[{name}] Bad Vertex Index: {vi}')
        res = C_ERROR
        
    # Check faces counts
    nbf3, nbf4, nbfo = 0,0,0
    for faceno in range(0,self.getNbFace()):
      #startIdx, endIdx = self.getFaceLastIdx(faceno), self.getFaceStartIdx(faceno)
      fao = self.matIdx[faceno]
      # Check Texture coordinates
      if fao.fAttr & C_FACE_TEXT:
        for vti in fao.tvertIdx: # [startIdx:endIdx]:
          if (vti<0) or (vti>=sizeTexList):
            print(f'Group[{name}].face[{faceno}] Bad Texture Index List: {vti}')
            res = C_ERROR

      if fao.fAttr & C_FACE_NORM:
        for vti in fao.normIdx: #[startIdx:endIdx]:
          if (vti<0) or (vti>=sizeNormList):
            print(f'Group[{name}].face[{faceno}] Bad Normal Index: {vti}')
            res = C_ERROR
      
      nbv = self.getFaceLastIdx(faceno) - self.getFaceStartIdx(faceno)
      if nbv==3: 
        nbf3+=1
      elif nbv==4:
        nbf4+=1
      else:
        nbfo+=1

    print(f'Group[{name}] has {nbf3} Triangles, {nbf4} Quad, {nbf4} others')
    
    return res
    
  # -----------------------------------------------------------------------------
  def calcXYRadius(self, lstVxIdx):
    ''' Compute the Maximum Radius on Oxy plan of a list of vertex indices fo this group. '''
    rmax = sys.float_info.min
  
    for p in [ self.coordList[i] for i in lstVxIdx ]:
      r = p.x*p.x + p.y*p.y
      if r>rmax:
        rmax = r
  
    return math.sqrt(rmax)
  
  # -----------------------------------------------------------------------------
  def removeFace(self, faceno=-1, materialName=None):
    ''' Remove a face from a group.
    Enhancement: a more pythonic way (instead the copy of the java code)
    
    Parameters
    ----------
    faceno : int
      Face to remove (if >0)

    materialName : str
      Name of the material. Remove the first face with this material

    Returns
    -------
    int
      C_OK : Success
      C_ERROR : faceno==-1 and materialName not found
    '''
    if faceno==-1:
      if materialName:
        try:
          matidx = self.lstMat.index(materialName)
          faceno = self.findFace(matidx)
          if faceno<0:
            return C_FAIL
        except ValueError:
          return C_FAIL
      else:
        return C_ERROR
      
    startidx = self.getFaceStartIdx(faceno)
    lastidx = self.getFaceLastIdx(faceno)
    delta = lastidx - startidx

    del self.matIdx[faceno]
    del self.vertIdx[startidx:lastidx]
      
    debsc = self.stripCount[:faceno] # if faceno>0 else [ ]
    finsc = array.array('l', [ sc-delta for sc in self.stripCount[faceno+1:] ])
    self.stripCount[:] = debsc + finsc
    
    return C_OK

  # -----------------------------------------------------------------------------
  def removeFaces(self, lstFaceNo=None, materialName=None):
    ''' Remove a set of faces given by their face numbers or a materialName. '''
    nbface = self.getNbFace()
    if not lstFaceNo:
      matidx = self.lstMat.index(materialName)
      lstFaceNo = [ fi for fi in range(0, nbface) if self.matIdx[fi].fAttr & C_FACE_MATMASK==matidx ]
      
    nstripcount = [] # array.array('l')
    nvertidx = [] # array.array('l')
    nmatIdx = [] # array.array('L')
      
    for fi in range(0,len(self.stripCount)-1):
      stidx = self.stripCount[fi]
      lastidx = self.stripCount[fi+1]
      
      if not fi in lstFaceNo:
        nstripcount.append(len(nvertidx))
        nmatIdx.append(self.matIdx[fi])
        nvertidx += self.vertIdx[stidx:lastidx]
            
    nstripcount.append(len(nvertidx))
    self.stripCount = array.array('l', nstripcount)
    self.matIdx = nmatIdx
    self.vertIdx = array.array('l', nvertidx)
  
  def extractFaces(self, destName=None, materialName=None, inGeom=True):
    ''' Extract from the current group, the faces that match the condition
    Create a new group
    if inGeom is False: Create a new geometry
    else add the new group to the geometry of the current group
    '''    
    wg = self.geom if inGeom else self.geom.copy()
    matidx, destIdx = -1,-1
    try:
      matidx = self.lstMat.index(materialName)
    except ValueError:
      logging.warning('Source Material[%s] not found', materialName)
      return None
    
    try:
      if destName:
        destIdx = self.lstMat.index(destName)
    except ValueError:
      logging.warning('Destination Material[%s] not found', destName)
      return None
      
    ngrp = wg.createGeomGroup(destName if destName else self.getName()+'_extracted_' + materialName)
    
    if inGeom:
      nStripCount = [0, ]
      nVertIdx = [ ]
      
      for faceno, fao in enumerate(self.matIdx):
        if fao.fAttr & C_FACE_MATMASK==matidx:
          nVertIdx += self.getFaceVertIdx(faceno)
          nStripCount.append(len(nVertIdx))
          nfao = FaceAttr(fao)
          if destName:
            nfao.setMatIdx(destIdx)
          ngrp.matIdx.append(nfao)
          
      ngrp.stripCount = array.array('l', nStripCount)
      ngrp.vertIdx = array.array('l', nVertIdx)
                
    else:
      print('extractFaces not inGeom : NOT IMPLEMENTED YET')
      return None
    
    # Remove extracted faces
    self.removeFaces(materialName=materialName)
    
    return ngrp
    
    
      
  
  def FaceFusion(self, prevMatName, newMatName):  
    ''' Fusion of the faces with prevMatName into a single face with newMatName.
    If the faces are not connex, it may generate several faces.
    Texture coordinates are lost
    '''
    #Ensure the new Material exists
    self.geom.addMaterial(newMatName)
    
    matidx = self.lstMat.index(prevMatName)
    lstFacesNo, lstFacesIdx = [ ], []
    for faceno,fao in enumerate(self.matIdx):
      #fao = self.matIdx[faceno]
      if fao.fAttr & C_FACE_MATMASK==matidx:
        lstFacesIdx.append(self.getFaceVertIdx(faceno))
        lstFacesNo.append(faceno) 
    
    if len(lstFacesNo)==1:
      # No Fusion required - Just change Mat name
      self.setMatIdx(lstFacesNo[0], self.lstMat.index(newMatName))
      return
    
    # Create a dictionnary of unoriented edges
    #   dict of [ idx0, idx1, List of nextEdgeIdx Ptr=[], used=False, code, count, tidx0, tidx1 ]
    edgeDict = { }
    bigFaceText = True

    for i,f in enumerate(lstFacesIdx):
      
      # Compute Best case for texture presence
      bigFaceText &= (self.matIdx[lstFacesNo[i]].fAttr & C_FACE_TEXT)!=0
      
      if bigFaceText:
        lt = self.getFaceTVertIdx(lstFacesNo[i])
        LstTegde = [ (lt[i], lt[i+1]) for i in range(0, len(lt)-1) ] + [ (lt[-1], lt[0]), ]
        
      # For Each Edge
      for edgeno, edge in enumerate([ (f[i], f[i+1]) for i in range(0, len(f)-1) ] + [ (f[-1], f[0]), ]):
        code = (min(edge) << 32) | max(edge)
        try:
          edgeDesc = edgeDict[code]
          edgeDesc[COUNT] += 1
        except KeyError:
          if bigFaceText:
            edgeDesc = [ edge[0], edge[1], [], False, code, 1, LstTegde[edgeno][0], LstTegde[edgeno][1] ]
          else:
            edgeDesc = [ edge[0], edge[1], [], False, code, 1 ]
            
          edgeDict[code] = edgeDesc
          
    bigFace = []
    
    # For Each Face
    for i,f in enumerate(lstFacesIdx):
      
      # For Each Edge
      for edge in [ (f[i], f[i+1]) for i in range(0, len(f)-1) ] + [ (f[-1], f[0]), ]:
        # if Edge is exterior (not shared) Add it to the new edge list
        code = (min(edge) << 32) | max(edge)
        edgeDesc = edgeDict[code]
        if edgeDesc[COUNT]==1:
          p0 = self.coordList[edge[0]]
          p1 = self.coordList[edge[1]]
          if bigFaceText: # Since bigFaceText is a best case, we add all or none texture coordinates
            p0.texture = self.texList[edgeDesc[6]]
            p1.texture = self.texList[edgeDesc[7]]
          bigFace.append( (p0, p1) )
    
    self.addFaceByEdges(bigFace, bigFaceText, newMatName)
    # Remove old faces
    self.removeFaces(materialName=prevMatName)
  
  


  # -----------------------------------------------------------------------------
  def fusion(self, angrp):
    ''' Fusion 'angrp' GeomGroup into self group.
    Both groups may belong to different WaveGeom

    Parameters
    ----------
    angrp : GeomGroup
      Group to add.
    '''
    logging.info("Start [%s:%d faces, %d Vx, %d Tx] <- [%s:%d faces, %d Vx, %d Tx]", self.getName(), self.getNbFace(), \
                 len(self.coordList), len(self.texList), \
                 angrp.getName(), angrp.getNbFace(), len(angrp.coordList),len(angrp.texList))
    
    if self.geom!=angrp.geom:
      # They belong to different WaveGeom
      # the destination WaveGeom must be upgraded with missing Vertex, TVertex and Norms
      # All indexes must me remapped
      
      # Copy the added group
      ngrp = GeomGroup(src=angrp)
      
      nbInit = len(ngrp.coordList)      
      nbMaxVerts = len(self.coordList)      

      if nbMaxVerts<1000: # and (nbMaxTVerts<1000):        
        # Case of empty dest list (append few times)
        if nbMaxVerts==0: # Dest group is empty : Vertex Mapping are already copied and correct
          logging.info("Direct Copy of %d vertex", len(ngrp.coordList))
          self.geom.coordList[:] = ngrp.geom.getCoordList()

          #logging.info("Direct Copy of %d Texture coords", len(ngrp.texList))
          self.geom.texList[:] = ngrp.geom.getTexList()

          #logging.info("Direct Copy of %d normals", len(ngrp.normList))
          self.geom.normList[:] = ngrp.geom.getNormList()
        else:
          logging.info("Using O(n2) method: %d", nbMaxVerts)
          # Create an empty mapping tables
          mapVert  = [-2] * nbInit
          mapTVert = [-2] * len(ngrp.texList)
          mapNorm  = [-2] * len(ngrp.normList)
          
          for i,vi in enumerate(ngrp.vertIdx):
            nvi = mapVert[vi]
            if nvi<0:
              nvi = IndexAdd(self.coordList, Point3d(ngrp.coordList[vi]))
              mapVert[vi] = nvi
              
            ngrp.vertIdx[i] = nvi

          for fao in ngrp.matIdx:
            if fao.fAttr & C_FACE_TEXT:
              for i,vti in enumerate(fao.tvertIdx):
                nvti = mapTVert[vti]
                if nvti<0:
                  nvti = IndexAdd(self.texList, TexCoord2f(ngrp.texList[vti]))
                  mapTVert[vti] = nvti
            
                fao.tvertIdx[i] = nvti
    
            if fao.fAttr & C_FACE_NORM:
              for i,vti in enumerate(fao.normIdx):
                nvti = mapNorm[vti]
                if nvti<0:
                  nvti = IndexAdd(self.normList, Vector3d(ngrp.normList[vti]))
                  mapNorm[vti] = nvti
            
                fao.normIdx[i] = nvti
      
      # TODO: Proceed with Lines
      else: # Use an octree and a PaveList2D to merge coordList(s)
        logging.info("Using Octree method: %d", nbMaxVerts)
                  # Create an empty mapping tables
        mapVert  = [-2] * nbInit
        mapTVert = [-2] * len(ngrp.texList)

        npTab = np.empty( (nbMaxVerts, 3) )
        for pNo, p in enumerate(self.coordList):
          npTab[pNo] = [ p.x, p.y, p.z ]
    
        # Create an KDTree with the 'known Vertex' in a "global" np.array
        tree = spatial.KDTree(npTab, leafsize=10 if nbMaxVerts<10000 else 100)
        svect = np.zeros((1,3))
        
        for i,vi in enumerate(ngrp.vertIdx):
          nvi = mapVert[vi]
          if nvi<0:
            # Search in KDTtree
            p = Point3d(ngrp.coordList[vi])
            svect[0] = [ p.x, p.y, p.z ]
            rest, resIdx = tree.query(svect)

            # if found (not too far) ==> Put it in a tmp table
            if rest[0]<FEPSILON: # Use an existing vertex
              nvi = int(resIdx[0])
            else: # Add a new vertex to the dest list of Vertex
              nvi = len(self.coordList)
              self.coordList.append(p)

            mapVert[vi] = nvi
          
          ngrp.vertIdx[i] = nvi
          
        #Remap Texture
        pl = PaveList2D(n=32, texList=self.texList)
        for fao in ngrp.matIdx:
          if fao.fAttr & C_FACE_TEXT:
            for i,vti in enumerate(fao.tvertIdx):
              nvti = mapTVert[vti]
              if nvti<0:
                texture = TexCoord2f(ngrp.texList[vti]) # FIX bad source texture index
                nvti = pl.IndexAdd(texture)
                mapTVert[vti] = nvti
                
              fao.tvertIdx[i] = nvti
          
      # Map Material indexes
      mapMatIdx = [ IndexAdd(self.lstMat, matName) for matName in ngrp.lstMat ]
      for fao in ngrp.matIdx:
        fao.fAttr = (fao.fAttr & C_FACE_TEXTNORM) | mapMatIdx[fao.fAttr & C_FACE_MATMASK]

      # Keep the new group for fusion
      angrp = ngrp
    # else VertIdx Maps are identity
        
    # Copy the faces 
    vertpos = len(self.vertIdx)
        
    self.matIdx   += angrp.matIdx
    self.vertIdx  += angrp.vertIdx
    self.stripCount = self.stripCount[:-1] + array.array('l', [ sc+vertpos for sc in angrp.stripCount ])
    logging.info("End Result [%s:%d faces, %d Vx]", self.getName(), self.getNbFace(), len(self.coordList))


  def linkCoordList(self, wg):
    ''' Ensure that the GeomGroup points to WaveGeom common data. '''
    self.coordList = wg.coordList if wg else None
    self.texList = wg.texList if wg else None
    self.normList = wg.normList if wg else None
    self.lstMat = wg.lstMat if wg else None
    self.geom = wg if wg else None


  
  
  #
  # Create a strip of faces with two loops in the geom group
  # l0 >= l1
  #
  #
  #     X---------------------X---------------------X
  #     |\                                    ----- |
  #     | \                               ----      |
  #     |  \                           ---          |
  #     |   x-------------------------x..           |
  #     |   |                         |  ...        |
  #     |   |                         |      ..     |
  #     |   |                         |        ...  |
  #     |   |                         |           ..X
  #     |   |                         |             |
  #     |   x--------e=l1.j0----------x..           |
  #     |  /                             ...        |
  #     | /                                  ..     |
  #     |/                                     .... |
  #     X---------------e0=l0.0---------------------X
  #
  # Return the list of faces defined by a list of vertex indexes
  #
  def createStrip(self, l0, l1, Rep0, Rep1, minAngle=0.7854, minCosinus=0.01):
    ''' Create a strip of faces from two loops in the geom group.

    Parameters
    ----------
    l0 : list of Edge
      First list of egdes

    l1 : list of Edge
      Second list of egdes

    Returns
    -------
    list of faces
        list of faces defined by a list of vertex indexes
    '''
  
    lstFaces = []
    nbl0 = len(l0)
    nbl1 = len(l1)

    if (nbl0==0) or (nbl1==0):
      return C_ERROR

    if nbl0<nbl1: # Swap lists and Coord Systems
      # The faces will be build in the wrong order (wrt normal)
      l0, nbl0, Rep0, l1, nbl1, Rep1 = l1, nbl1, Rep1, l0, nbl0, Rep0
      swaped = True
    else:
      swaped = False
    
    if WFBasic.PYPOS3D_TRACE:
      for i in range(0,nbl0):self.geom.addMaterial(f'DebugMat{i}')
      self.geom.curGroup = self
    
    hasTexture = min( [ l.hasTexture for l in l0 ] ) and min( [ l.hasTexture for l in l1 ] )
    
    # Change coordinates to l0 coordinate system
    l0 = Rep0.To(l0, hasTexture=hasTexture)
    l1 = Rep0.To(l1, hasTexture=hasTexture)
      
    # j0 : Edge number in l1 where p0 is the closest to the first point of l0
    dmin = sys.float_info.max
    P0 = l0[0].p0
    j0 = -1
    # New heuristic: Estimate the distance at the Middle of segment
    mide0 = Point3d(P0).add(l0[0].p1).scale(0.5)
    for j,e in enumerate(l1):
      d = mide0.distance(Point3d(e.p0).add(e.p1).scale(0.5))
      if d<dmin:
        j0 = j
        dmin = d
      
    oneAgain = True
    j, jmax = j0, (j0-1)%nbl1
    i, imax = 0, nbl0-1
    e0Consumed, e1Consumed = False, False
      
    # Plot the strips
    #Plot(l0,l1)
    
    nbFaceInit = self.getNbFace()
    
    while oneAgain:
      if WFBasic.PYPOS3D_TRACE:
        self.geom.selectMaterial(f'DebugMat{i}')
          
      e0 = l0[i]
      e1 = l1[j]

      configUtil = 2*e0Consumed + e1Consumed
      if configUtil==0: # 2 New edges
        # Try a Quadrangle
        amin, aidx, cmin, cidx, vn = Regularity(e0.p0, e0.p1, e1.p1, e1.p0)
        if WFBasic.PYPOS3D_TRACE: print('createStrip({0:d},{1:d}) - sin[{2:g}, {3:d}] - cos[{4:g}, {5:d}] - {6:s}'.format(i,j, amin*180.0/math.pi, aidx, cmin, cidx,str(vn)))

        if (cmin>minCosinus) and (amin>minAngle): # Create a QuadAngle
          # Create a quandrangular face          
          texIdxs = [ IndexAdd(self.texList, P.texture) for P in [ e0.p0, e0.p1, e1.p1, e1.p0 ] ] if hasTexture else [ ]
          vexIdxs = [e0.idx0, e0.idx1, e1.idx1, e1.idx0]
          e0Consumed = True
          e1Consumed = True
          if WFBasic.PYPOS3D_TRACE: print('  Quad {0:s} {1:s} {2:s} {3:s}'.format(str(e0.p0), str(e0.p1), str(e1.p1), str(e1.p0)))

        elif ((aidx==0) or (aidx==2)) or ((cidx==1) or (cidx==3)): # Create a T1 triangle
          texIdxs = [ IndexAdd(self.texList, P.texture) for P in [ e0.p0, e0.p1, e1.p0 ] ] if hasTexture else [ ]
          vexIdxs = [e0.idx0, e0.idx1, e1.idx0]
          e0Consumed = True
          e1Consumed = False
          if WFBasic.PYPOS3D_TRACE: print('  Tri1 {0:s} {1:s} {2:s}'.format(str(e0.p0), str(e0.p1), str(e1.p0)))

        else: # Create a T2 triangle
          texIdxs = [ IndexAdd(self.texList, P.texture) for P in [ e0.p0, e1.p1, e1.p0 ] ] if hasTexture else [ ]
          vexIdxs = [e0.idx0, e1.idx1, e1.idx0]
          e0Consumed = False
          e1Consumed = True          
          if WFBasic.PYPOS3D_TRACE: print('  Tri2 {0:s} {1:s} {2:s}'.format(str(e0.p0), str(e1.p1), str(e1.p0)))

      elif configUtil==1: # 1 New edge : e0
        texIdxs = [ IndexAdd(self.texList, P.texture) for P in [ e0.p0, e0.p1, e1.p1 ] ] if hasTexture else [ ]
        vexIdxs = [e0.idx0, e0.idx1, e1.idx1 ]
        e0Consumed = True

      elif configUtil==2: # 1 New edge : e1
        texIdxs = [ IndexAdd(self.texList, P.texture) for P in [ e0.p1, e1.p1, e1.p0 ] ] if hasTexture else [ ]
        vexIdxs = [e0.idx1, e1.idx1, e1.idx0]
        e1Consumed = True

      #else: print('Impossible') : e0Consumed and e0Consumed can be both false


      # Add the new face taking into accound the reverted loops
      if not swaped:
        vexIdxs.reverse()
        texIdxs.reverse()

      lstFaces += self.addFace( vexIdxs, texIdxs, [ ] )

      # Compute new egdes
      if (i==imax) and (j==jmax):
        if e0Consumed and (i!=imax):
          i+=1
          e0 = l0[i]
          e0Consumed = False
          
        if e1Consumed and (j!=jmax):
          j = (j+1)%nbl1 
          e1 = l1[j]
          e1Consumed = False
          
        oneAgain = False
      else:
        if e0Consumed and (i!=imax): 
          i+=1
          e0Consumed = False
              
        if e1Consumed and (j!=jmax):
          j = (j+1)%nbl1
          e1Consumed = False
    # Wend        
  
    # Final Edges may not be consumed
    # Create a last triangle
    if (not e0Consumed and e1Consumed):
      texIdxs = [ IndexAdd(self.texList, P.texture) for P in [ e0.p0, e0.p1, e1.p1 ] ] if hasTexture else [ ]
      vexIdxs = [e0.idx0, e0.idx1, e1.idx1 ]
      if swaped:
        vexIdxs.reverse()
        texIdxs.reverse()
      lstFaces += self.addFace( vexIdxs, texIdxs, [ ] )

    elif not e1Consumed and e0Consumed:
      texIdxs = [ IndexAdd(self.texList, P.texture) for P in [ e0.p1, e1.p1, e1.p0 ] ] if hasTexture else [ ]
      vexIdxs = [e0.idx1, e1.idx1, e1.idx0]
      if swaped:
        vexIdxs.reverse()
        texIdxs.reverse()
      lstFaces += self.addFace( vexIdxs, texIdxs, [ ] )

    if WFBasic.PYPOS3D_TRACE: print('==> createStrip[{0:s}]: InitFace={1:d} NewFaces={2:d}'.format(self.getName(), nbFaceInit, self.getNbFace()))
      
    return lstFaces    
       
       
  def calcCoordSyst(self, faceno=0, orientation='XZY'):
    ''' Create a coordinate system from the first (default) face of a group.
    The face is supposed to be a square (at least) 4 edges

    (Oy)

     3        2
      +-------+
      ^       |
   eu |   X   |
      |       |
      +------>+          (Ox)
     0   ev    1

    Parameters
    ----------
    faceno : int
        Index of the face to use
    orientation : str
      Orientation of coord (NOT USED - RESERVED FOR FUTURE)

    Returns
    -------
    CoordSyst
      A new 3D Coordinate System at the center of the face (X, eu, ev)
      the third vector of the coordinate system: ew = eu x ev
    '''

    vxtab = self.getFaceVertex(faceno)
    eu = Vector3d( vxtab[3].x - vxtab[0].x, vxtab[3].y - vxtab[0].y, vxtab[3].z - vxtab[0].z).normalize()
    ev = Vector3d( vxtab[1].x - vxtab[0].x, vxtab[1].y - vxtab[0].y, vxtab[1].z - vxtab[0].z).normalize()       
    center = Point3d(vxtab[3].x + vxtab[1].x, vxtab[3].y + vxtab[1].y, vxtab[3].z + vxtab[1].z).scale(0.5)
    
    return CoordSyst(center, eu, ev)
  
  def calcBarycentre(self):
    '''Compute and return the iso barycentre of the used vertex. '''
    # Consider each vertex once
    return Point3d.barycentre([ self.coordList[i] for i in set(self.vertIdx) ])

  # ---------------------------------------------------------------------------
  # TODO: Move it in WFBasic for a more general usage and wall it the vertex
  # list of the GeomGroup
  #
  def RadialScale(self, center, eu, ev, dh, ds, R, nbLoop=0, topPlaneGrp=None):
    ''' Perform a quadratic Radial Scaling of vertexes.

    - Modified Point are below the top Coord. Syst and z>0.0 in the (center, eu, ev) one
      The top Coord system is either given by topPlaneGrp of computed as follow:
      TopCoord = ( center + (0,0,dh), eu, ev )
    - Modify the impunt vertex
    - The scaling curve is defined by the point [-ds, 0] and [+dh, R] 
      where R is a supposed Radius of the vextex at coordinate z = dh

      The Parabol z = a.r² 
      

      O = the center
   
                                ^ Scale Ratio
                                |                           .....  X (usually 1.0 (100%))
                                |                .....   X
                                |        ..... X
                                |    ..X
                                |  .. 
                               .x..
                             .  |
                           .    |
                          .     |
                         .      |
      ------------------x-------O-----------------------------------|--------------> Oz
                       -ds      |                                   dh
                                |        .....
 

 
    Parameters
    ----------
    center : Point3d
      Center of the bottom Plan

    eu : Vector3d
      First Vector of the bottom Plan

    ev : Vector3d
      Second Vector of the bottom Plan (eu and ev are supposed unitary and orthog)

    dh : float
      Height limit of the Parabol
    ds : float
      Down limit of the Parabol
    R : float
      Radius of the parabol in dh

    Returns
    -------
    GeomGroup
      returns self
  
    '''
    if WFBasic.PYPOS3D_TRACE: print('RadialScale({0:s}) ------------------------ '.format(self.getName()))
  
    # Compute the Cutting axis and the transformation matrix of the reference plan
    ew,M,MT = CreateTransform(eu, ev)
  
    # Compute the 'a' coef of the Parabol z = a.r² - ds
    a = (dh+ds) / R / R
  
    rd2 = R*R
  
    if topPlaneGrp:
      centerTop, euTop, evTop = topPlaneGrp.calcCoordSyst()

      ewTop, Mtop, MTtop = CreateTransform(euTop, evTop)

      # For each point in the upper cylinder and below the topPlane : Do a radial quadratic scaling
      for p in self.geom.coordList:
        
        # In top the Coord Syst
        pInTop = Point3d(p).sub(centerTop).inLin33(Mtop)
        
        if pInTop.z<FEPSILON:
          # Change for Cutting Plan Coordinate System
          p.sub(center).inLin33(M)
          
          # if (p.z>=-FEPSILON) and (p.z <= dh+FEPSILON) and (p.x*p.x+p.y*p.y < rd2): FIX Avoid double z condition
          if (p.z>=-FEPSILON) and (p.x*p.x+p.y*p.y < rd2):
            k = math.sqrt( (p.z + ds) / a) / R
            p.scale(k)
      
          # Change (back) to initial Coordinate System
          p.inLin33(MT).add(center) 
      
    else:
      # For each point in the upper cylinder : Do a radial quadratic scaling
      for p in self.geom.coordList:
        
        # Change for Cutting Plan Coordinate System
        p.sub(center).inLin33(M)
        
        if (p.z>=-FEPSILON) and (p.z <= dh+FEPSILON) and (p.x*p.x+p.y*p.y < rd2):
          k = math.sqrt( (p.z + ds) / a) / R
          p.scale(k)
    
        # Change (back) to initial Coordinate System
        p.inLin33(MT).add(center) 
    
    return self


  def FaceRadialScale(self, center, eu, ev, pc=1.0, lstFaceNo=None, materialName=None):
    ''' Perform a radial scaling of the vertex of a list of faces.
    '''
    if WFBasic.PYPOS3D_TRACE: print('FaceRadialScale({0:s}) ------------------------ '.format(self.getName()))
    
    nbface = self.getNbFace()
    if not lstFaceNo:
      matidx = self.lstMat.index(materialName)
      lstFaceNo = [ fi for fi in range(0, nbface) if self.matIdx[fi].fAttr&C_FACE_MATMASK==matidx ]
 
    # Compute the Cutting axis and the transformation matrix of the reference plan
    ew,M,MT = CreateTransform(eu, ev)
    
    ptMap = { vi : self.coordList[vi] for fi in lstFaceNo for vi in self.vertIdx[self.stripCount[fi]:self.stripCount[fi+1]] }
        
    # For each point in the upper cylinder : Do a radial quadratic scaling
    for p in ptMap.values():
      # Change for Cutting Plan Coordinate System
      p.sub(center).inLin33(M)
      
      p.scale(pc)
  
      # Change (back) to initial Coordinate System
      p.inLin33(MT).add(center) 
    
    return self
  

  def translate(self, tx=0.0, ty=0.0, tz=0.0):
    ''' Translate the group of the given vector.
    This method creates new vertex (Point3d) in the coordList (usually in WaveGeom)
    
    Parameters
    ----------
    tx,ty,tz : Translation coordinates
    
    Return
    ------
    self for convenience
    '''
    # Create an empty mapping tables
    nbInit = len(self.coordList)      
    mapVert  = [-2] * nbInit

    vertIdxSet = set( self.vertIdx )
    
    v = Point3d(tx,ty,tz)
    nbPt = nbInit
    # Create new translated points
    for vi in vertIdxSet:
      p = Point3d(self.coordList[vi])
      p.add(v)
      self.coordList.append(p)
      mapVert[vi] = nbPt
      nbPt += 1
      
    # Remap Group's vertex indexes with the new ones
    self.vertIdx[:] = array.array('l', [ mapVert[vi] for vi in self.vertIdx ] )
    
    return self
    
  
  def rotate(self, rx=0.0, ry=0.0, rz=0.0, cx=0.0, cy=0.0, cz=0.0, AxisOrder='xyz'):
    ''' Rotate the group around standard axis with an optional rotation center and
    according given rotation orders.    
    Angles in degrees.
    This method creates new vertex (Point3d) in the coordList (usually in WaveGeom)
    
    Parameters
    ----------
    rx,ry,rz : Axis rotation angles (float)
    cx,cy,cz : Center coordinates
    axisOrder : str, defaul 'xyz'
      Rotation order between 'xyz', 'yzx' and 'zxy'
    
    Return
    ------
    self for convenience
    
    '''
    # Create an empty mapping tables
    nbInit = len(self.coordList)      
    mapVert  = [-2] * nbInit

    vertIdxSet = set( self.vertIdx )
    
    # Create the center for a simple origin usage
    center = Point3d(cx,cy,cz)

    theta = rx*math.pi/180.0
    c, s = math.cos(theta), math.sin(theta)
    MRotX = [ \
       [ 1.0, 0.0, 0.0], \
       [ 0.0,   c, - s ], \
       [ 0.0,   s,   c ], \
       ]
  
    theta = ry*math.pi/180.0
    c, s = math.cos(theta), math.sin(theta)
    MRotY = [ \
       [   c, 0.0,   s ], \
       [ 0.0, 1.0, 0.0 ], \
       [ - s, 0.0,   c ], \
       ]

    theta = rz*math.pi/180.0
    c, s = math.cos(theta), math.sin(theta)
    MRotZ = [ \
      [   c, - s, 0.0], \
      [   s,   c, 0.0 ], \
      [ 0.0, 0.0, 1.0 ], \
      ]

    if AxisOrder=='xyz':
      m = MxProd(MRotY, MRotX)
      m = MxProd(MRotZ, m)
    elif AxisOrder=='yzx':
      m = MxProd(MRotZ, MRotY)
      m = MxProd(MRotX, m)
    else: # zxy
      m = MxProd(MRotX, MRotZ)
      m = MxProd(MRotY, m)
    
    nbPt = nbInit
    # Create new rotated points
    for vi in vertIdxSet:
      p = Point3d(self.coordList[vi])
      p.sub(center)
      p.inLin33(m)
      p.add(center)
      self.coordList.append(p)
      mapVert[vi] = nbPt
      nbPt += 1
      
    # Remap Group's vertex indexes with the new ones
    self.vertIdx[:] = array.array('l', [ mapVert[vi] for vi in self.vertIdx ] )
    
    return self
    
  def symetry(self, cx=0.0, cy=0.0, cz=0.0, sym='symxy'):
    ''' Mirror the group along standard planes with an optional origin center.

    This method creates new vertex (Point3d) in the coordList (usually in WaveGeom)
    
    Parameters
    ----------
    cx,cy,cz : Center coordinates
    sym : str, defaul 'symxy'
      Plan for symetry 'symxy', 'symyz' and 'symzx'
    
    Return
    ------
    self for convenience
    '''
    # Create an empty mapping tables
    nbInit = len(self.coordList)      
    mapVert  = [-2] * nbInit

    vertIdxSet = set( self.vertIdx )

    nbPt = nbInit
    # Create the center for a simple origin usage
    c = Point3d(cx,cy,cz)
    if sym=='symxy':
      # Create new symetrized points
      for vi in vertIdxSet:
        p = Point3d(self.coordList[vi])
        p.z = 2.0*c.z - p.z
        self.coordList.append(p)
        mapVert[vi] = nbPt
        nbPt += 1
      
    elif sym=='symyz':
      # Create new symetrized points
      for vi in vertIdxSet:
        p = Point3d(self.coordList[vi])
        p.sub(c)
        p.x = 2.0*c.x - p.x
        p.add(c)
        self.coordList.append(p)
        mapVert[vi] = nbPt
        nbPt += 1
    else: # symzx
      # Create new symetrized points
      for vi in vertIdxSet:
        p = Point3d(self.coordList[vi])
        p.y = 2.0 * c.y - p.y 
        self.coordList.append(p)
        mapVert[vi] = nbPt
        nbPt += 1
    
      
    # Remap Group's vertex indexes with the new ones
    self.vertIdx[:] = array.array('l', [ mapVert[vi] for vi in self.vertIdx ] )
    
    # Invert face orders
    self.invertFaceOrder()
    
    return self
    
    
  def calcFaceOrientation(self, faceno=-1, materialName=None, AxisOrder='xyz'):
    ''' Compute the orientation of a face according to an Axis order (xyz, yzx, zxy).
    
    The orientation is the set of rotations that aligns the first axis of the
    axis order on the face normal vector.
    
    Parameters
    ----------
    faceno : int
      Face to remove (if >0)

    materialName : str, default None
      Name of the material. Use the first face with this material
      
    axisOrder : str, default 'xyz'
      Rotation order between 'xyz', 'yzx' and 'zxy'

    Returns
    -------
    int, Vector3d
      C_OK : Success
      C_FAIL : faceno==-1 and materialName not found
      C_ERROR : Face Index Error
    '''
    if faceno==-1:
      if materialName:
        try:
          matidx = self.lstMat.index(materialName)
          faceno = self.findFace(matidx)
          if faceno<0:
            return C_FAIL, None
        except ValueError:
          return C_FAIL, None
      else:
        return C_ERROR, None
    elif (faceno<0) or (faceno>=self.getNbFace()):
      return C_ERROR, None

    ret, vr = C_OK, None
    vxtab = self.getFaceVertex(faceno)    
    if vxtab and len(vxtab)>=3:
      # Perform the orientation computation  
      vr = Point3d.face_orientation(vxtab, AxisOrder=AxisOrder, inDegrees=True)              
    else:  
      ret = C_FAIL
            
    return ret, vr


  def unTriangularize(self, maxsin=FEPSILON, algo=C_FACE_NONE, surfaceThreshold=0.1):
    ''' Untriangularize the current group (inplace).
    Triangular faces are merged when:
    * They have the same material
    * They have a common edge
    * The sinus of their normals (angle) is smaller than 'maxsin' and 
      when 'algo' parameter is C_FACE_SURF if the merged face represents more than 'surfaceThreshold' percent
      of the final face.
    
    Parameters
    ----------
    maxsin: float, Optional, default = FEPSILON
      Maximal value for the sinus of the angle of both
      normals to permit a face merge
      
    algo: bitmask for algorithm selection, default = C_FACE_NONE
      Add C_FACE_SURF for relative surface filtering. 
      
      Add either C_FACE_ORDER_LENGTH, C_FACE_ORDER_ANGLE or C_FACE_ORDER_SURF for surface choice ordering
      The ordering is used to choose the merged triangle between the 3 neighbours of a destination triangle.
      C_FACE_ORDER_LENGTH: The triangle with the greastest common edge is merged
      C_FACE_ORDER_ANGLE: The triangle with the smallest angle is merged
      C_FACE_ORDER_SURF: The triangle with the greatest relative surface is merged
      
    surfaceThreshold: float, default=0.1
      Threshold used by the surface filtering to keep a triangle as a merging candidate.
      The candidat is kept is its surface will represent more than 'surfaceThreshold' of the
      merged quad surface.
      
    Return
    ------
    int : 
      >=0    : Number of faces in the result group
      C_FAIL : No face to merge
    
    '''
    lstTri = []
    nVertIdx = [] 
    nMatIdx = [] 
    nStrip = [0, ]
    initNbFace = self.getNbFace()
    
    # Initialize a list of list of triangles indexed by the vertex Id
    vertexId2Tri = [ None ] * len(self.coordList)
    
    for noface in range(0, initNbFace):
       
      startIdx = self.getFaceStartIdx(noface)
      lastIdx = self.getFaceLastIdx(noface)
      keepIt = (lastIdx-startIdx!=3)
      
      if not keepIt: # It's a triangle
        # Create a triangular face with some pre-computed values
        p0,p1,p2 = self.vertIdx[startIdx], self.vertIdx[startIdx+1], self.vertIdx[startIdx+2]
        tf = TriFace(noface, p0, p1, p2, coordList=self.coordList)

        if tf.surface > FEPSILON2: # Filter too small or null surface
          lstTri.append(tf)
          
          for i in [p0, p1, p2]:
            l = vertexId2Tri[i]
            if l is None:
              vertexId2Tri[i] = [ tf, ]
            elif not tf in l:
              l.append(tf)
        else: # Triangular surface is too small, Store it
          keepIt = True
           
      if keepIt: # Just copy non triangular face OR rejected triangles
        nVertIdx += self.vertIdx[startIdx:lastIdx]
        nStrip.append(len(nVertIdx))
        nMatIdx.append(self.matIdx[noface])
    
    dn2 = maxsin*maxsin
    # Check triangle by 2 with a filter on the material index and according to the 'algo'
    # a comparison of surfaces
    while lstTri:
      # Get the first triangle
      curTri = lstTri.pop()
      p0, p1, p2 = curTri.p0, curTri.p1, curTri.p2
      ltri0, ltri1, ltri2 = vertexId2Tri[p0], vertexId2Tri[p1], vertexId2Tri[p2]

      ltri0.remove(curTri)
      ltri1.remove(curTri)
      ltri2.remove(curTri)
      
      fao = self.matIdx[curTri.noface]
      mi = fao.fAttr
      
      # Compute the list of candidate triangles:
      lstCandidateTri = ltri0.copy() if ltri0 else []
      lstCandidateTri += [ t for t in (ltri1 if ltri1 else []) if t not in lstCandidateTri ]
      lstCandidateTri += [ t for t in (ltri2 if ltri2 else []) if t not in lstCandidateTri ]
      
      # Compute the list of adjacent triangles (up to 3)
      e0,e1,e2 = curTri.edges
      lstAdj = [ t for t in lstCandidateTri \
                  if (self.matIdx[t.noface].fAttr==mi) and \
                    ( (e0 in t.revedges) or (e1 in t.revedges) or (e2 in t.revedges) ) and \
                    (curTri.vnorm.cross(t.vnorm).norme2() <= dn2) ]
      
      if algo & C_FACE_SURF: # Add a filter on surface comparison
        lstAdj = [ t for t in lstAdj if math.fabs(t.surface/(curTri.surface + t.surface)) >= surfaceThreshold ]
        
      if lstAdj: # Choose the 'best' triangle to merge
        adjTri, ceCurTri = curTri.choose(lstAdj, algo)

        # Remove Adjacent Triangle from the list of available faces
        lstTri.remove(adjTri)
        
        vertexId2Tri[adjTri.p0].remove(adjTri)
        vertexId2Tri[adjTri.p1].remove(adjTri)
        vertexId2Tri[adjTri.p2].remove(adjTri)
          
        # Merge curTri and adjTri
        cea = adjTri.commEdge(curTri)
        pp = adjTri.p2 if cea==0 else (adjTri.p0 if cea==1 else adjTri.p1)
        if ceCurTri==0:
          tv = [ curTri.p0, pp, curTri.p1, curTri.p2 ]
        elif ceCurTri==1:
          tv = [ curTri.p0, curTri.p1, pp, curTri.p2 ]
        else:
          tv = [ curTri.p0, curTri.p1, curTri.p2, pp ]

        nVertIdx += tv
        nStrip.append(len(nVertIdx))
        nMatIdx.append(fao)
        
        adjFao = self.matIdx[adjTri.noface]
        
        if fao.fAttr & C_FACE_TEXT:
          tp = adjFao.tvertIdx[2] if cea==0 else (adjFao.tvertIdx[0] if cea==1 else adjFao.tvertIdx[1])

          if ceCurTri==0:
            fao.tvertIdx = array.array('l', [ fao.tvertIdx[0], tp, fao.tvertIdx[0+1], fao.tvertIdx[0+2] ])
          elif ceCurTri==1:
            fao.tvertIdx = array.array('l', [ fao.tvertIdx[0], fao.tvertIdx[0+1], tp, fao.tvertIdx[0+2] ])
          else:
            fao.tvertIdx = array.array('l', [ fao.tvertIdx[0], fao.tvertIdx[0+1], fao.tvertIdx[0+2], tp ])

        if self.matIdx[curTri.noface].fAttr & C_FACE_NORM: 
          np = adjFao.normIdx[2] if cea==0 else (adjFao.normIdx[0] if cea==1 else adjFao.normIdx[1])

          if ceCurTri==0:
            fao.normIdx = array.array('l', [ fao.normIdx[0], np, fao.normIdx[0+1], fao.normIdx[0+2] ])
          elif ceCurTri==1:
            fao.normIdx = array.array('l', [ fao.normIdx[0], fao.normIdx[0+1], np, fao.normIdx[0+2] ])
          else:
            fao.normIdx = array.array('l', [ fao.normIdx[0], fao.normIdx[0+1], fao.normIdx[0+2], np ])
        
      else: # The current triangle has no relevant adjacent, Add it to the new group
        startIdx = self.getFaceStartIdx(curTri.noface)
        lastIdx = self.getFaceLastIdx(curTri.noface)
        nVertIdx += self.vertIdx[startIdx:lastIdx]
        nStrip.append(len(nVertIdx))
        nMatIdx.append(self.matIdx[curTri.noface])
      
    #End of While LstTri
    
    # Finish the modification of the group
    self.vertIdx[:] = array.array('l', nVertIdx) 
    self.stripCount[:] = array.array('l', nStrip)
    self.matIdx[:] = nMatIdx
    
    return self.getNbFace() if initNbFace > self.getNbFace() else C_FAIL
    
      
  
# Named tuple replaced by a 'sloted' class for both performance and extension
#TriFace = collections.namedtuple('TriFace', ['noface', 'p0', 'p1', 'p2', 'vnorm', 'edges', 'revedges'])

class TriFace:
  ''' This class represents a triangular face used during 'Untriangularization' algorithm.
  '''
  noface:int
  p0:int
  p1:int
  p2:int
  edges:tuple
  revedges:tuple
  vnorm:float
  lenedges:tuple
  surface:float
  
  def __init__(self, noface, p0, p1, p2, coordList=None):
    
    self.noface = noface
    self.p0 = p0          
    self.p1 = p1
    self.p2 = p2
              
    self.edges =  ( (p0,p1), (p1,p2), (p2,p0) )
    self.revedges = ( (p0,p2), (p2,p1), (p1,p0) )
    
    # Perform computations
    if coordList:
      cp0,cp1,cp2 = coordList[p0], coordList[p1], coordList[p2]
      
      # Compute Normale
      self.vnorm = Point3d.triangle_normal(cp0, cp1, cp2)
      
      # Compute each edges length
      l0,l1,l2 = cp0.distance(cp1), cp1.distance(cp2), cp2.distance(cp0)
      self.lenedges = ( l0, l1, l2 )
      
      # Compute Surface (Heron formula)
      # Protect from rounding effects that create very small negative s2
      p = 0.5*(l0+l1+l2)
      s2 = p*(p-l0)*(p-l1)*(p-l2)
      self.surface = 0.0 if s2 < 0.0 else math.sqrt(s2) 
      
  def commEdge(self, adjt):
    return 0 if (self.p0,self.p1) in adjt.revedges else \
           (1 if (self.p1,self.p2) in adjt.revedges else \
            (2 if (self.p2, self.p0) in adjt.revedges else -1))
           
  def choose(self, lstAdj, algo):
    adjTri = lstAdj[0]
    if algo & C_FACE_ORDER_ANGLE:
      sinus2 = self.vnorm.cross(adjTri.vnorm).norme2()
      
      if len(lstAdj)>1:
        ns = self.vnorm.cross(lstAdj[1].vnorm).norme2()
        if ns < sinus2:
          adjTri = lstAdj[1]
          sinus2 = ns
          
      if len(lstAdj)>2:
        ns = self.vnorm.cross(lstAdj[2].vnorm).norme2()
        if ns < sinus2:
          adjTri = lstAdj[2]

      ceCurTri = self.commEdge(adjTri)
      
    elif algo & C_FACE_ORDER_SURF: #Keep the most contributing triangle
      deltas = math.fabs( adjTri.surface / (adjTri.surface + self.surface ) )
      
      if len(lstAdj)>1:
        nd = math.fabs( lstAdj[1].surface / (lstAdj[1].surface + self.surface))
        if nd > deltas:
          adjTri = lstAdj[1]
          deltas= nd
          
      if len(lstAdj)>2:
        nd = math.fabs( lstAdj[2].surface / (lstAdj[2].surface + self.surface))
        if nd > deltas:
          adjTri = lstAdj[2]

      ceCurTri = self.commEdge(adjTri)

    else: # algo & C_FACE_ORDER_LENGTH:    
      # Find the common Edge
      ceCurTri = self.commEdge(adjTri)
      lenEdge  = self.lenedges[ceCurTri]
      
      if len(lstAdj)>1:
        ce = self.commEdge(lstAdj[1])
        le = self.lenedges[ce]
        if le > lenEdge:
          adjTri = lstAdj[1]
          lenEdge = le
          ceCurTri = ce
          
      if len(lstAdj)>2:
        ce = self.commEdge(lstAdj[2])
        le = self.lenedges[ce]
        if le > lenEdge:
          adjTri = lstAdj[2]
          ceCurTri = ce
      
      
    return adjTri, ceCurTri
#











# -*- coding: utf-8 -*-
'''
Created on 4 oct. 2020
Setup of library delivery
'''
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pypos3d", # Replace with your own username
    version="1.3.1",
    author="Olivier Dufailly",
    author_email="dufgrinder@laposte.net",
    description="Wavefront files and Poser files manipulation library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://sourceforge.net/projects/pojamas",
    
    package_dir={'': 'src'},  
    packages=setuptools.find_packages(where='src', include=('pypos3d', 'pypos3d.*', 'langutil' )),

    #include_package_data=True,
    #package_data={ 'pypos3d':[ '../CHANGELOG.md', '../LICENSE'], },
    
    classifiers=[
        'Development Status :: 4 - Beta',
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    install_requires=['numpy>=1.19', 'scipy>=1.5', 'xlrd>=1.2'],
    python_requires='>=3.6',
)



# -*- coding: utf-8 -*-
'''  
Created on 3 oct. 2004
Ported in Python in May 2020
'''
import sys
import logging
import copy
from collections import namedtuple

import numpy as np
from scipy import spatial

from langutil import C_OK, C_FAIL, C_ERROR
from pypos3d.wftk.WFBasic import Vector3d, Point3d, BoundingBox, BoundingSphere, calcMLS
from pypos3d.wftk.Repere import Repere
from pypos3d.wftk.PoserFileParser import PoserFileParser, ParsingErrorException
from pypos3d.wftk.WaveGeom import WFMat, WFColor
from pypos3d.pftk.PoserBasic import PoserConst, PoserObject, PToken, PoserToken, Lang, TAS, index, create, WBoolLine, RemoveQuotes, nodeNameNo
from pypos3d.pftk.SimpleAttribut import KSA, ValueOpDelta, OffSA, TextureMapSA, SimpleAttribut

 
PtMapping = namedtuple('PtMapping', ['srcNo', 'refNo', 'refGeom', 'dist' ])
  
#   
class StructuredAttribut(PoserObject):
  ''' Structured attribute of a Poser file : NAME { list of PoserObject } '''
 
  def __init__(self, n=''):
    super(StructuredAttribut, self).__init__()
    self.setName(n)
    self._lstAttr = [ ]

  def getLstAttr(self): return self._lstAttr

  # 
  # Read a structured attribute data from the file. This class name, the name
  # and the opening bracket are supposed to be consumed by the caller.
  # Because it is the mean it uses to recognize a structured attribute.
  #    
  def read(self, st):    
    code,cn,rw = st.getLine()
    
    while (code!=PoserFileParser.TT_EOF) and (code!=PoserFileParser.TT_RIGHTBRACKET):
      if st.ttype == PoserFileParser.TT_WORD:
        try:
          vc = Lang[cn]
          #  Known word
          if vc.isStructured:
            sta = create(vc, rw)

            self.addAttribut(sta)

            #  Read the opening bracket
            code,cn,rw = st.getLine()
            if code==PoserFileParser.TT_LEFTBACKET:
              sta.read(st)
            else:
              if sta.isAmbi():
                st.pushBack()
              else:
                logging.warning( "Line[%s]: '{' is missing for %s", st.lineno, cn)
          else:
            if vc.isDirect:
              self.setDirect(vc, rw)
            else:
              sa = create(vc, cn)
              #  Read before add
              sa.read(st, rw)
              self.addAttribut(sa)

        except KeyError: #  Mot inconnu
          logging.info("Line[%s] - Unknown word:%s", st.lineno(), cn)

      #  Get next line
      code,cn,rw = st.getLine()
      # Wend

  def isAmbi(self): return self.ambiguious

  # @see deyme.v3d.poser.PoserObject#write(java.io.FileWriter)
  def write(self, fw, pfx):
    nextPfx = pfx + "\t"
    fw.write(f"{pfx}{self.getPoserType().token} {self.getName()}\n{nextPfx}{{\n")
    for po in self._lstAttr:
      po.write(fw, nextPfx)

    if not self._lstAttr:
      fw.write('\n')

    fw.write(f"{nextPfx}}}\n")

  def getAttribut(self, attrTypeOrName):
    if isinstance(attrTypeOrName, PToken):
      po = next((x for x in self._lstAttr if x.getPoserType() == attrTypeOrName), None)
    else:
      po = next((x for x in self._lstAttr if x.getName() == attrTypeOrName), None)

    return po
    #  20080417 : Fix - Shall return null if not found

# Useless as we always use less 'free' attributs
#   def replaceAttribut(self, attrTypeName, nattr):
#     for i,po in enumerate(self._lstAttr):
#       if po.getPoserType()==attrTypeName:
#         self._lstAttr[i] = nattr
#         return True
# 
#     return False

  def deleteAttribut(self, attrName):
    for po in self._lstAttr:
      if po.getName() == attrName:
        self._lstAttr.remove(po)
        return True

    return False

  def addAttribut(self, po):
    self._lstAttr.append(po)

  def setDirect(self, tokenID, val):
    #  Dummy (debug) implementation
    logging.warning("Unexpected Direct:" + tokenID.token + " (" + (val if val else 'null') + ") in " + str(self.__class__) + " for " + self.getPoserType().token)
    # Workaround: Create a simple atttribut and inster it 'as usual'
    sa = SimpleAttribut(tokenID, val)
    self._lstAttr.append(sa)

  def findAttribut(self, attrTypeName:PoserToken, name=None) -> PoserObject :
    if name:
      try:
        po = next(x for x in self._lstAttr if (x.getPoserType() == attrTypeName) and (x.getName() == name))
        return po
      except:
        #  20080417 : Fix - Shall return null if not found
        return None
    else:
      return self.getAttribut(attrTypeName)

#TODO 
#  * This class represents a set of vertice indexes used by 
#  * Poser to adapt dynamic clothes (I think). 
#  * Moreover it can contain embedded attribute like 'stitchVertsGroupProperties'
#  
class VertsGroup(StructuredAttribut):
  def __init__(self):
    super(VertsGroup, self).__init__()
    self.vertList = None
    self.vertTab = []

  #    * @see deyme.v3d.poser.PoserObject#read(deyme.v3d.poser.PoserFileParser)
  #    
  def read(self, st):
    self.vertTab = [ ]

    fin = False
    while not fin:
      code,cn,rw = st.getLine()
      if code == PoserFileParser.TT_WORD:
        if cn == "v":          
          self.vertTab.append(int(rw))
          continue 

        #  Other keyword found
        try:
          vc = Lang[cn]
          #  Known word
          if vc.isStructured:
            sta = create(vc, rw)
            self.addAttribut(sta)
            #  Read the opening bracket
            code,cn,rw = st.getLine()
            if code==PoserFileParser.TT_LEFTBACKET:
              sta.read(st)
            else:
              if sta.isAmbi():
                st.pushBack()
              else:
                logging.warning("Line[%s]:  '{' is missing for %s", st.lineno(), cn)
          else:
            sa = create(vc, cn)
            #  Read before add
            sa.read(st, rw)
            self.addAttribut(sa)

        except:
          #  Mot inconnu
          logging.info("Line[%s] - Unknown word:%s",st.lineno(),  cn)
      elif (code==PoserFileParser.TT_EOF) or (code==PoserFileParser.TT_RIGHTBRACKET): 
        break
      else:
        logging.warning("Line[%s] - Not Accepted :%s", st.lineno(), cn)
        raise ParsingErrorException()

    # End of read

  #    * @see deyme.v3d.poser.PoserObject#write(java.io.FileWriter)
  def write(self, fw, pfx):
    #  FIXME: Wrong output - Other attributes (like stitchVertsGroupProperties are forgotten)
    nextPfx = pfx + "\t"
    fw.write(f"{pfx}{self.getPoserType().token} {self.getName()}\n{nextPfx}{{\n")
    nextPfx = pfx + "\tv "

    for vi in self.vertTab:
      fw.write(f"{nextPfx}{vi}\n")
 
    fw.write(f"{nextPfx}}}\n")

# 
# This class represents the "textureFile" element of Poser files.
#  
class TextureFile(StructuredAttribut):

  def __init__(self):
    super(TextureFile, self).__init__()
    self._file = '' #  "NO_MAP" 
    self._warped = False

  def setDirect(self, tokenID, val):
    if tokenID==PoserToken.E_file:
      self._file = None if val == PoserConst.C_NO_MAPG or val == PoserConst.C_NO_MAP else val
    elif tokenID==PoserToken.E_warped:
      self._warped = val.startswith("1")
    else:
      logging.warning("Unexpected Direct:%s (%s)",tokenID.token, val)

  #  (non-Javadoc)
  #    * @see deyme.v3d.poser.PoserObject#write(java.io.FileWriter)
  #    
  def write(self, fw, pfx):
    fw.write(f"{pfx}{self.getPoserType().token} {self.getName()}\n")
    nextPfx = pfx + "\t"
    fw.write(f"{nextPfx}{{\n")
    fw.write(f"{nextPfx}file {PoserConst.C_NO_MAPG if self._file==None else self._file}\n") #  "NO_MAP" 
    WBoolLine(fw, nextPfx, "warped", self._warped )
    for po in self._lstAttr:
      po.write(fw, nextPfx)
    if not self._lstAttr:
      fw.write('\n')
    fw.write(f"{nextPfx}}}\n")

 

# 
# This class represents the "nodeInput" element of Poser files.
#  
class NodeInput(StructuredAttribut):
  def __init__(self, nodeName='', colorStr='1 1 1', src=None):
    super(NodeInput, self).__init__()
    self._poserType = PoserToken.E_nodeInput
    
    if src:
      self._nodeName = src._nodeName
      self._name = nodeName if nodeName else src._name
      self._value    = src._value
      self._parmR    = src._parmR
      self._parmG    = src._parmG
      self._parmB    = src._parmB
      self._node     = src._node
      self._file     = src._file 
    else:
      self._nodeName = nodeName
      if nodeName:
        self._name = nodeName
        
      self._value    = colorStr #  22 0 32768
      self._parmR    = "" #  NO_PARM
      self._parmG    = "" #  NO_PARM
      self._parmB    = "" #  NO_PARM
      self._node     = "" #  NO_NODE
      self._file     = "" #  "NO_MAP" 
    
  # Set the name.
  def setName(self, s):
    self._name = RemoveQuotes(s)
    self._nodeName = self._name

  def read(self, st):
       
    while True:
      code,cn,val = st.getLine()
      if code==PoserFileParser.TT_WORD:
        try:
          tokenID = Lang[cn]        
        
          if tokenID==PoserToken.E_name:
            self._nodeName = RemoveQuotes(val)
          elif tokenID==PoserToken.E_value:
            self._value = val
          elif tokenID==PoserToken.E_parmR:
            self._parmR = None if val==PoserConst.C_NO_PARM else val
          elif tokenID==PoserToken.E_parmG:
            self._parmG = None if val==PoserConst.C_NO_PARM else val
          elif tokenID==PoserToken.E_parmB:
            self._parmB = None if val==PoserConst.C_NO_PARM else val
          elif tokenID==PoserToken.E_node:
            self._node = None if val==PoserConst.C_NO_NODE else RemoveQuotes(val)
          elif tokenID==PoserToken.E_file:
            val = val.strip()
            self._file = None if val==PoserConst.C_NO_MAPG or val==PoserConst.C_NO_MAP else val
          else:
            logging.warning("Unexpected Direct in NodeInput:%s (%s)",tokenID.token, val)
        
        except KeyError: #  Mot inconnu
          logging.info("Line[%s] - Unknown word:%s", st.lineno(), cn)
      elif (code==PoserFileParser.TT_RIGHTBRACKET) or (code==PoserFileParser.TT_EOF): 
        break
      else:
        logging.warning("L[%d] - Not Accepted :%s", st.lineno(), cn)
        raise ParsingErrorException()




  #    * @see deyme.v3d.poser.PoserObject#write(java.io.FileWriter)
  #    
  def write(self, fw, pfx):
    fw.write(f"{pfx}{self.getPoserType().token} \"{self.getName()}\"\n")
    nextPfx = pfx + "\t"
    fw.write(f"{nextPfx}{{\n")
    fw.write(f"{nextPfx}name \"{self._nodeName}\"\n")
    fw.write(f"{nextPfx}value {self._value}\n")
    fw.write(f"{nextPfx}parmR {self._parmR if self._parmR else PoserConst.C_NO_PARM}\n")
    fw.write(f"{nextPfx}parmG {self._parmG if self._parmG else PoserConst.C_NO_PARM}\n")
    fw.write(f"{nextPfx}parmB {self._parmB if self._parmB else PoserConst.C_NO_PARM}\n")
    fw.write(f"{nextPfx}node \"{self._node if self._node else PoserConst.C_NO_NODE}\"\n")
    fw.write(f"{nextPfx}file {self._file if self._file else PoserConst.C_NO_MAPG}\n")
    fw.write(f"{nextPfx}}}\n")




  def getFile(self): return self._file

  #
  # Mimic Poser Python interface
  # Set the _file with a filename value
  #
  def setString(self, fn):
    self._file = fn

  #
  # Mimic Poser Python interface
  # Set the _parmR with a float value
  #
  def setFloat(self, f):
    t = self._value.split(' ')
    self._value = str(f) + ' ' + t[1] + ' ' + t[2]

  def setParmR(self, s):
    self._parmR = s

  def getInNode(self, sht):
    if self._node and self._node!=PoserConst.C_NO_NODE:
      return sht.getNodeByInternalName(self._node)
    return None

#
#  Shader Tree Node
#
class Node(StructuredAttribut):
  def __init__(self, src=None):
    super(Node, self).__init__()
    self._poserType = PoserToken.E_node
    self.ambiguious = True
    
    # Copy constructor
    if src:
      self._name      = src._name
      self._nodeName  = src._nodeName
      self._typeName  = src._typeName
      self._pos       = src._pos
      self._collapsed = src._collapsed
      self._showPreview = src._showPreview
      lsrc = src.getLstAttr()
      if lsrc:
        for ni in lsrc:
          nci = NodeInput(src=ni)
          self.addAttribut(nci)
          
    else:
      self._nodeName  = ''
      self._typeName  = ''
      self._pos       = '0 0'
      self._collapsed = False
      self._showPreview = False

  def setDirect(self, tokenID, val):
    if tokenID == PoserToken.E_name:
      self._nodeName = RemoveQuotes(val)
    elif tokenID==PoserToken.E_inputsCollapsed:
      self._collapsed = (val[0] == "1")
    elif tokenID==PoserToken.E_showPreview:
      self._showPreview = (val[0] == "1")
    else:
      logging.warning("Unexpected Direct in Node:%s (%s)",tokenID.token, val)
  
  def addAttribut(self, po):
    if po.getPoserType()==PoserToken.E_pos:
      self._pos = po.getValue()
    else:
      super(Node, self).addAttribut(po)

  # Set the name.
  def setName(self, s):
    if s:
      t = s.split(' ')
      self._name = RemoveQuotes(t[1])
      self._typeName = RemoveQuotes(t[0])
    else:
      self._name = s
      
  def setCollapsed(self, inputsCollapsed):
    self._collapsed = inputsCollapsed
    
  def setShowPreview(self, showPreview):
    self._showPreview = showPreview

  # @see deyme.v3d.poser.PoserObject#write(java.io.FileWriter)
  def write(self, fw, pfx):
    fw.write(f"{pfx}{self.getPoserType().token} \"{self._typeName}\" \"{self.getName()}\"\n")
    if self._lstAttr:
      nextPfx = pfx + "\t"
      fw.write(f"{nextPfx}{{\n")
      fw.write(f"{nextPfx}name \"{self._nodeName}\"\n")
      fw.write(f"{nextPfx}pos {self._pos}\n")

      if self._collapsed:
        WBoolLine(fw, nextPfx, "inputsCollapsed", self._collapsed)
      if self._showPreview:
        WBoolLine(fw, nextPfx, "showPreview", self._showPreview)
        
      for po in self._lstAttr:
        po.write(fw, nextPfx)
        
      fw.write(f"{nextPfx}}}\n")

  def getInputByInternalName(self, inpName='Blending'):
    return next( (n for n in self.getLstAttr() if isinstance(n, NodeInput) and n._nodeName==inpName), None)

#   def getInputByName(self, name):
#     return next( (n for n in self.getLstAttr() if isinstance(n, NodeInput) and n.getName()==name), None)


  def setLocation(self, x, y):
    self._pos = str(x)+ ' ' + str(y)

  def getNodeType(self):
    return PoserConst.kNodeTypeCodeBLENDER if self._typeName=='blender' else PoserConst.kNodeTypeCodeIMAGEMAP if self._typeName=='image_map' else C_ERROR

#
# Shader Tree
# 
class ShaderTree(StructuredAttribut):
  def __init__(self):
    super(ShaderTree, self).__init__()
    self.setPoserType(PoserToken.E_shaderTree)

  def getNodeByInternalName(self, intName):
    return next( (n for n in self.getLstAttr() if isinstance(n, Node) and n._nodeName==intName), None)

  def getNodeByName(self, name):
    return next( (n for n in self.getLstAttr() if isinstance(n, Node) and n.getName()==name), None)


  # Create a Node in a shader tree
  # Partial Implementation for 'blender' and 'image_map'
  #
  # Pos as (x,y)
  # Default Blender:
  #			node "blender" "Blender"
  #				{
  #				name "Blender"
  #				pos 396 54
  #       inputsCollapsed 1
  #       showPreview 1
  #				nodeInput "Input_1"
  #					{
  #					name "Input_1"
  #					value 1 1 1
  #					parmR NO_PARM
  #					parmG NO_PARM
  #					parmB NO_PARM
  #					node NO_NODE
  #					file NO_MAP
  #					}
  #				nodeInput "Input_2"
  #					{
  #					name "Input_2"
  #					value 1 1 1
  #					parmR NO_PARM
  #					parmG NO_PARM
  #					parmB NO_PARM
  #					node NO_NODE
  #					file NO_MAP
  #					}
  #				nodeInput "Blending"
  #					{
  #					name "Blending"
  #					value 0.5 0 1
  #					parmR NO_PARM
  #					parmG NO_PARM
  #					parmB NO_PARM
  #					node NO_NODE
  #					file NO_MAP
  #					}  
  def CreateNode(self, poserkNodeTypeCode, pos=None, inputsCollapsed=False, showPreview=False): # kNodeTypeCodeBLENDER or kNodeTypeCodeIMAGEMAP
    n = Node()
    # Generated a name according to the type
    baseName = 'Blender_' if poserkNodeTypeCode==PoserConst.kNodeTypeCodeBLENDER else 'Image_Map_'

    lstno = [ nodeNameNo(no._nodeName) for no in self.getLstAttr() \
      if no.getNodeType()==poserkNodeTypeCode ]

    no = max(lstno) + 1 if lstno else 0

    n._name = baseName+str(no)
    n._nodeName = n._name
    n._typeName = 'blender' if poserkNodeTypeCode==PoserConst.kNodeTypeCodeBLENDER else 'image_map'

    # Add the relevant NodeInput to the new Node
    if poserkNodeTypeCode==PoserConst.kNodeTypeCodeBLENDER:
      lstni = (("Input_1", "1 1 1"), ("Input_2", "1 1 1"), ("Blending", "0 0 1"))
    else:
      lstni = (("Image_Source", "-1 0 0"), ("Auto_Fit", "0 0 0"), ("U_Scale", "1 -1 1"),\
                  ("V_Scale", "1 -1 1"), ( "U_Offset", "0 0 1"), ("V_Offset", "0 0 1"), \
                  ("Texture_Coords", "1 0 0"), ("Image_Mapped", "4 0 0"), ("Background", "1 1 1"),\
		  ("Global_Coordinates", "0 0 1"), ("Mirror_U", "0 0 0"), ("Mirror_V", "0 0 0"), \
                  ("Texture_Strength", "1 0 1"), ("Filtering", "3 0 0"))

    n._lstAttr += [ NodeInput(tni[0], colorStr=tni[1]) for tni in lstni ]
    self._lstAttr.append(n)

    if pos:
      n.setLocation(pos[0], pos[1])
    
    n.setCollapsed(inputsCollapsed)
    n.setShowPreview(showPreview)
    return n
 
  def AttachTreeNodes(self, nodeDest, inputName, nodeRef):
    ni = nodeDest.getInputByInternalName(inpName=inputName)
    ni._node = nodeRef.getName()


#  
# material class.  
#  
class PoserMaterial(StructuredAttribut):
  # 
  # Create the default "skin" material.
  #
  def __init__(self, src=None):
    super(PoserMaterial, self).__init__()
    
    # Create the Minimal Direct attributs
    self._KdColor = WFColor(0.992157, 0.992157, 0.992157, 1.) 
    self._KaColor = WFColor(0., 0., 0., 1.) 
    self._KsColor = WFColor(0.733333, 0.733333, 0.733333, 1.) 
    self._TextureColor = WFColor(1.0, 1.0, 1.0, 1.0) 
    self._NsExponent = 10 
    self._tMin = 0.
    self._tMax  = 0. 
    self._tExpo = 0.6 
    self._bumpStrength = 1.
    self._ksIgnoreTexture = 0 
    self._reflectThruLights = 1 
    self._reflectThruKd = 0. 
    self._textureMap = TextureMapSA(PoserToken.E_textureMap) 
    self._bumpMap = TextureMapSA(PoserToken.E_bumpMap)   # NO_MAP
    self._reflectionMap = TextureMapSA(PoserToken.E_reflectionMap)  # NO_MAP
    self._transparencyMap = TextureMapSA(PoserToken.E_transparencyMap)   # NO_MAP
    self._ReflectionColor = WFColor(0.733333, 0.733333, 0.733333, 1.) 
    self._reflectionStrength = 1. 
    self._shaderTree = None
    self.setPoserType(PoserToken.E_material) # FIX: Shall be set anyway
    
    if src:
      if isinstance(src, WFMat):
        # Create a Material from a simple WaveFront material
        self.setName(src.name)
        self._KdColor = WFColor(src.kd[0], src.kd[1], src.kd[2], src.kd[3])
        self._KaColor = WFColor(src.ka[0], src.ka[1], src.ka[2], src.ka[3])
        self._KsColor = WFColor(src.ks[0], src.ks[1], src.ks[2], src.ks[3])
        
        self._textureMap = TextureMapSA(PoserToken.E_textureMap, initPath=src.map_kd)
        
      elif isinstance(src, PoserMaterial):
        
        self.setName(src.getName())
        self._KdColor =  WFColor(src._KdColor.r, src._KdColor.g, src._KdColor.b, src._KdColor.t)
        self._KaColor =  WFColor(src._KaColor.r, src._KaColor.g, src._KaColor.b, src._KaColor.t)
        self._KsColor =  WFColor(src._KsColor.r, src._KsColor.g, src._KsColor.b, src._KsColor.t)
        self._TextureColor =  WFColor(src._TextureColor.r, src._TextureColor.g, src._TextureColor.b, src._TextureColor.t)
        self._NsExponent =  src._NsExponent 
        self._tMin = src._tMin 
        self._tMax  =  src._tMax  
        self._tExpo =  src._tExpo 
        self._bumpStrength = src._bumpStrength 
        self._ksIgnoreTexture =  src._ksIgnoreTexture 
        self._reflectThruLights =  src._reflectThruLights 
        self._reflectThruKd =  src._reflectThruKd 
        self._textureMap = TextureMapSA(initPath=src._textureMap) 
        self._bumpMap = TextureMapSA(initPath=src._bumpMap)
        self._reflectionMap = TextureMapSA(initPath=src._reflectionMap)
        self._transparencyMap = TextureMapSA(initPath=src._transparencyMap)
        self._ReflectionColor = WFColor(src._ReflectionColor.r, src._ReflectionColor.g, src._ReflectionColor.b, src._ReflectionColor.t)
        self._reflectionStrength = src._reflectionStrength
        if src._shaderTree:
          self._shaderTree = ShaderTree()
          
          for n in src._shaderTree.getLstAttr():
            # It is supposed that ShaderTree only contains Nodes
            nn = Node(src=n)
            self._shaderTree.addAttribut(nn)
          
          
      else: # Illegal source
        raise TypeError() 

  def getTextureMap(self): return self._textureMap.getPath() #if self._textureMap else None 

  def getBumpMap(self): return self._bumpMap.getPath() #if self._bumpMap else None

  def getReflectionMap(self): return self._reflectionMap.getPath() #if self._reflectionMap else None

  def getTransparencyMap(self): return self._transparencyMap.getPath() #if self._transparencyMap else None

  def getShaderTree(self): return self._shaderTree

  def getLstNodes(self):
    st = self._shaderTree
    return [ po for po in st.getLstAttr() if isinstance(po, Node) ] if st else [ ]

  def setDirect(self, tokenID, val):
    if tokenID == PoserToken.E_KdColor:
      tv = val.split()
      self._KdColor = WFColor(float(tv[0]), float(tv[1]), float(tv[2]), float(tv[3]))
      
    elif tokenID == PoserToken.E_KaColor:
      tv = val.split()
      self._KaColor = WFColor(float(tv[0]), float(tv[1]), float(tv[2]), float(tv[3]))
      
    elif tokenID == PoserToken.E_KsColor:
      tv = val.split()
      self._KsColor = WFColor(float(tv[0]), float(tv[1]), float(tv[2]), float(tv[3]))
       
    elif tokenID == PoserToken.E_TextureColor:
      tv = val.split()
      self._TextureColor = WFColor(float(tv[0]), float(tv[1]), float(tv[2]), float(tv[3]))
      
    elif tokenID == PoserToken.E_NsExponent:
      self._NsExponent = float(val)
      
    elif tokenID == PoserToken.E_tMin:
      self._tMin = float(val)
      
    elif tokenID == PoserToken.E_tMax:
      self._tMax  = float(val)
      
    elif tokenID == PoserToken.E_tExpo:
      self._tExpo = float(val)
      
    elif tokenID == PoserToken.E_bumpStrength:
      self._bumpStrength = float(val)
      
    elif tokenID == PoserToken.E_ksIgnoreTexture:
      self._ksIgnoreTexture = int(val)
      
    elif tokenID == PoserToken.E_reflectThruLights:
      self._reflectThruLights = int(val)
      
    elif tokenID == PoserToken.E_reflectThruKd:
      self._reflectThruKd = float(val)
      
    elif tokenID == PoserToken.E_ReflectionColor:
      tv = val.split()
      self._ReflectionColor = WFColor(float(tv[0]), float(tv[1]), float(tv[2]), float(tv[3]))
       
    elif tokenID == PoserToken.E_reflectionStrength:
      self._reflectionStrength = float(val)

    else:
      logging.warning("Unexpected Direct in Material:%s (%s)",tokenID.token, val)
  
  def addAttribut(self, po):
    tokenID = po.getPoserType()
    if tokenID == PoserToken.E_textureMap:
      self._textureMap = po
       
    elif tokenID == PoserToken.E_bumpMap:
      self._bumpMap = po
      
    elif tokenID == PoserToken.E_reflectionMap:
      self._reflectionMap = po
      
    elif tokenID == PoserToken.E_transparencyMap:
      self._transparencyMap = po
    
    elif tokenID==PoserToken.E_shaderTree:
      self._shaderTree = po
    else:
      super(PoserMaterial, self).addAttribut(po)

  # @see deyme.v3d.poser.PoserObject#write(java.io.FileWriter)
  def write(self, fw, pfx):
    fw.write(f"{pfx}{self.getPoserType().token} {self.getName()}\n")
    nextPfx = pfx + "\t"
    fw.write(f"{nextPfx}{{\n")
    fw.write(f"{nextPfx}KdColor {self._KdColor.r} {self._KdColor.g} {self._KdColor.b} {self._KdColor.t}\n") 
    fw.write(f"{nextPfx}KaColor {self._KaColor.r} {self._KaColor.g} {self._KaColor.b} {self._KaColor.t}\n")
    fw.write(f"{nextPfx}KsColor {self._KsColor.r} {self._KsColor.g} {self._KsColor.b} {self._KsColor.t}\n")
    fw.write(f"{nextPfx}TextureColor {self._TextureColor.r} {self._TextureColor.g} {self._TextureColor.b} {self._TextureColor.t}\n") 

    fw.write(f"{nextPfx}NsExponent {self._NsExponent}\n") 
    fw.write(f"{nextPfx}tMin {self._tMin}\n") 
    fw.write(f"{nextPfx}tMax {self._tMax}\n") 
    fw.write(f"{nextPfx}tExpo {self._tExpo}\n") 
    fw.write(f"{nextPfx}bumpStrength {self._bumpStrength}\n") 
    fw.write(f"{nextPfx}ksIgnoreTexture {self._ksIgnoreTexture}\n") 
    fw.write(f"{nextPfx}reflectThruLights {self._reflectThruLights}\n") 
    fw.write(f"{nextPfx}reflectThruKd {self._reflectThruKd}\n") 
    
    self._textureMap.write(fw, nextPfx)
    self._bumpMap.write(fw, nextPfx)
    self._reflectionMap.write(fw, nextPfx)
    self._transparencyMap.write(fw, nextPfx)

    fw.write(f"{nextPfx}ReflectionColor {self._ReflectionColor.r} {self._ReflectionColor.g} {self._ReflectionColor.b} {self._ReflectionColor.t}\n") 
    fw.write(f"{nextPfx}reflectionStrength {self._reflectionStrength}\n") 

    if self._shaderTree:
      self._shaderTree.write(fw, nextPfx)

    if self._lstAttr:
      for po in self._lstAttr:
        po.write(fw, nextPfx)

    fw.write(f"{nextPfx}}}\n")


# 
# Representation of an Inverse Kynematic Chain
#  
class InkyChain(StructuredAttribut):

  # on
  def __init__(self):
    super(InkyChain, self).__init__()
    self._pmName = None
    self._active = False

  # 
  #    * Return true if the actor or the prop is ON
  #    * @return a boolean
  #    
  def isActive(self): return self._active

  def setDirect(self, tokenID, val):
    if tokenID==PoserToken.E_name:
      self._pmName = val
    elif tokenID==PoserToken.E_on:
      self._active = True
    elif tokenID==PoserToken.E_off:
      self._active = False
    else:
      logging.warning("Unexpected Direct in InkyChain:%s (%s)", tokenID.token, val)

  # 
  #    * @see pftk.poser.kern.StructuredAttribut#write(java.io.PrintWriter,
  #    
  def write(self, fw, pfx):
    fw.write(f"{pfx}{self.getPoserType().token} {self.getName()}\n")
    nextPfx = pfx + "\t"
    fw.write(f"{nextPfx}{{\n")
    fw.write(f"{nextPfx}name {self._pmName}\n")
    fw.write(f"{nextPfx}{'on' if self._active else 'off'}\n")
    for po in self._lstAttr:
      po.write(fw, nextPfx)
    if not self._lstAttr:
      fw.write('\n')
    fw.write(f"{nextPfx}}}\n")

# 
#  * This class represents the "keys" entry of every channel definition.
#  * 
#  * keys
#  *  {
#  *  static <0|1>
#  *  k <long> <real>
#  *  [sl <0|1>
#  *  spl|lin|con
#  *  sm|br]
#  *  }
#  * 
#  * static determines whether this channel is animating (=0) or static (=1).
#  * k denotes the key information. The first values is the frame, starting at 0. The second value is the channel value at this frame.
#  * sl determines whether or not to loop animation.
#  * spl|lin|con denote the section interpolation type: spline, linear, constant.
#  * sm|br represent either a smooth transition or break in the transition.
#  
class Keys(StructuredAttribut):

  def __init__(self, src=None):
    super(Keys, self).__init__()
    self.setPoserType(PoserToken.E_keys)
    self._static = src._static if src else False
    self._hshKeys = { frameNo:KSA(src=ksa) for frameNo,ksa in src._hshKeys.items() } if src else { } # Replaced by a dict frameNo:KSA

  def read(self, st):
    curK = None
    while True:
      code,cn,rw = st.getLine()
      if code==PoserFileParser.TT_WORD:
        try:
          vc = Lang[cn]

          if vc==PoserToken.E_k:
            tv = rw.split()
            curK = KSA(int(tv[0]), float(tv[1]))
            self._hshKeys[curK._noFrame] = curK
            # Read 3 lines after (if any)
            c0, _, cval = st.getLine()
            
            if c0==PoserFileParser.TT_RIGHTBRACKET: 
              break
            
            _, cty, *_ = st.getLine()
            _, ccnx, *_ = st.getLine()

            # elif vc==PoserToken.E_sl:
            curK.setSl(cval[0]=='1')
            
            # elif (vc==PoserToken.E_spl) or (vc==PoserToken.E_lin) or (vc==PoserToken.E_con):
            curK.setCurveType(Lang[cty])

            #elif (vc==PoserToken.E_sm) or (vc==PoserToken.E_br):
            curK.setCurveCnx(Lang[ccnx])
            
          elif vc==PoserToken.E_static:
            self._static = (rw[0]=='1')
          else:
            logging.warning("L[%d] - Not Accepted :%s", st.lineno(), cn)
            
        except KeyError:
          pass
      elif (code==PoserFileParser.TT_RIGHTBRACKET) or (code==PoserFileParser.TT_EOF): 
        break
      else:
        logging.warning("L[%d] - Not Accepted :%s", st.lineno(), cn)
        raise ParsingErrorException()


  def write(self, fw, pfx):
    fw.write(f"{pfx}{self.getPoserType().token}\n")
    nextPfx = pfx + "\t"
    fw.write(f"{nextPfx}{{\n")
    WBoolLine(fw, nextPfx, "static", self._static )
    for ks in sorted(self._hshKeys.values(), key=lambda k:k._noFrame):
      ks.write(fw, nextPfx)
    fw.write(f"{nextPfx}}}\n")

  def setStatic(self, s):
    self._static = s

  def addKey(self, noFrame, factor):
    self._hshKeys[noFrame] = KSA(noFrame, factor)

  # 
  # Return the K factor for a given frameNo.
  # Intermediate values are not yet calculated.
  #  
  # @param frameNo 0 based frame index
  # @return the found key factor. 0.0 if not found.
  #    
  def getKeyFactor(self, frameNo):
    try:
      return self._hshKeys[frameNo]._factor
    except KeyError:
      return 0.0

  # 
  #    * Set the K factor for a given frameNo.
  #    * Intermediate values are not yet calculated.
  #    * 
  #    * @param frameNo 0 based frame index
  #    * @return C_OK when the frame no exists and C_FAIL if not found.
  #    
  def setKeyFactor(self, frameNo, k):
    try:
      self._hshKeys[frameNo].setFactor(k)
      return C_OK
    except KeyError:
      return C_FAIL

  # 
  #  Return the list of declared "k" frames.
  #    
  # def getLstKey(self): return self._lstKeys

  # 
  # Convert a GenericTransform to a static one (i.e. no keys)
  # @return    The number of deleted keys
  #    
  def toStatic(self, staticValue):
    l = len(self._hshKeys)
    self._static = True
    #  Remove all keys
    self._hshKeys.clear()
    cle = KSA(0, staticValue)
    cle.setSl(True)
    cle.setCurveType(PoserToken.E_spl)
    cle.setCurveCnx(PoserToken.E_sm)
    self._hshKeys[0] = cle
    return l - 1

class HairGrowthGroup(StructuredAttribut):
  pass

class DocDescription(StructuredAttribut):
  def __init__(self):
    super(DocDescription, self).__init__()
    self._dimensions = None
    self._screenPlace = None
    self._displayMode = None

  #  USEPARENT
  def setDirect(self, tokenID, val):
    if tokenID==PoserToken.E_dimensions:
      self._dimensions = val
    elif tokenID==PoserToken.E_screenPlace:
      self._screenPlace = val
    elif tokenID==PoserToken.E_displayMode:
      self._displayMode = val
    else:
      logging.warning("Unexpected Direct in DocDescription:%s (%s)", tokenID.token, val)

  #    * @see deyme.v3d.poser.PoserObject#write(java.io.FileWriter)
  def write(self, fw, pfx):
    fw.write(f"{pfx}{self.getPoserType().token}\n")
    nextPfx = pfx + "\t"
    fw.write(f"{nextPfx}{{\n")
    if self._dimensions:
      fw.write(f"{nextPfx}dimensions {self._dimensions}\n")

    if self._screenPlace:
      fw.write(f"{nextPfx}screenPlace {self._screenPlace}\n")

    if self._displayMode:
      fw.write(f"{nextPfx}displayMode {self._displayMode}\n")

    for po in self._lstAttr:
      po.write(fw, nextPfx)

    fw.write(f"{nextPfx}}}\n")

# Description part of a figure.
#  class FigureDescription(StructuredAttribut): --> inlined in Figure
        

class AlternateGeom(StructuredAttribut):
  def __init__(self):
    super(AlternateGeom, self).__init__()
    #  Index of the Alternate geometry in the list (not a real attribut) 
    self._no = 0
    self._printName = None

    #  Geom filename formated like Poser with ':' or ""  
    self._geomFileName = None

  def setDirect(self, tokenID, val):
    if tokenID==PoserToken.E_name:
      self._printName = val
    elif tokenID==PoserToken.E_objFile:      
      bci1 = val.rfind(' ')
      self._geomFileName = val[(bci1 + 1):]
    else:
      logging.warning("Unexpected Direct in AlternateGeom:%s (%s)", tokenID.token, val)

  #  @see pftk.poser.kern.StructuredAttribut#write(java.io.PrintWriter,
  def write(self, fw, pfx):
    fw.write(f"{pfx}{self.getPoserType().token} {self.getName()}\n")
    nextPfx = pfx + "\t"
    fw.write(f"{nextPfx}{{\n")
    fw.write(f"{nextPfx}name {self._printName}\n")
    fw.write(f"{nextPfx}objFile 20 {self._geomFileName}\n")
    fw.write(f"{nextPfx}}}\n")


  def setNo(self, n):
    self._no = n

  def getNo(self): return self._no

  # Set the print name 
  #  @param gfn
  def setPrintName(self, gfn):
    self._printName = gfn

  # Set the filename as represented on the "objFile" line.
  #  @param gfn
  def setGeomFileName(self, gfn):
    self._geomFileName = gfn

  def getGeomFileName(self): return self._geomFileName


# Represents a delta for the vertex identified by noPt.
class DeltaPoint(Vector3d):

  def __init__(self, no=0, dx=0.0, dy=0.0, dz=0.0):
    super(DeltaPoint, self).__init__(dx,dy,dz)
    self.noPt = no

  def getPointNo(self): return self.noPt

  def setPointNo(self, n):
    self.noPt = n

  def __str__(self):
    return "d {0:d} {1: 11.8f} {2: 11.8f} {3: 11.8f}".format(self.noPt, self.x, self.y, self.z)

  # Round the value to C_MIN_DELTA
  #@classmethod --> Replaced by the buildin round(f,ndigit)
  #def round(cls, v):

  def getVector(self): return self

  def setVector(self, v):
    self.x = v.x
    self.y = v.y
    self.z = v.z

  def toV3d(self): return Vector3d(self)

# 
# This class represents a Deltas attribute used to morph a meshed object. 
#  
class Deltas(StructuredAttribut):

  def __init__(self, src=None):
    super(Deltas, self).__init__()
    self.setPoserType(PoserToken.E_deltas)

    #In python deltaTab replaced by a dict { noPt:DeltaPoint() }
    self.deltaSet = { no:DeltaPoint(dp.noPt, dp.x, dp.y, dp.z) for no,dp in src.deltaSet.items() } if src else { }

  #    * @see deyme.v3d.poser.PoserObject#read(deyme.v3d.poser.PoserFileParser)
  def read(self, st):
    self.deltaSet = { }

    while True:
      code,cn,rw = st.getLine()
      if (code==PoserFileParser.TT_WORD) and (cn=="d"):
        tv = rw.split()
        p = DeltaPoint(int(tv[0]), float(tv[1]), float(tv[2]), float(tv[3]))
        self.deltaSet[p.noPt] = p
      elif (code==PoserFileParser.TT_RIGHTBRACKET) or (code==PoserFileParser.TT_EOF): 
        break
      else:
        logging.warning("L[%d] - Not Accepted :%s", st.lineno(), cn)
        raise ParsingErrorException()

  # @see deyme.v3d.poser.PoserObject#write(java.io.FileWriter)
  def write(self, fw, pfx):
    fw.write(f"{pfx}{self.getPoserType().token}\n")
    nextPfx = pfx + "\t"
    fw.write(f"{nextPfx}{{\n")
    for d in sorted(self.deltaSet.values(), key=lambda p:p.noPt):
      fw.write(f"{nextPfx}{d}\n")
    fw.write(f"{nextPfx}}}\n")

  def clear(self):
    self.deltaSet.clear() # = { }

  @classmethod
  def enhancement(cls, srcWG, setFoundDeltas, setNewDeltas, enhance, boundingType, useVicinityLoop, alpha):
    res = C_OK
    mlsFailed = False

    # Populate result with extact matches
    setNewDeltas.update(setFoundDeltas)

    if enhance != PoserConst.C_NO_ENHANCEMENT:
      # 03FEV2008 : Check For missed/not complete faces 
      for grp in srcWG.getGroups():

        # Create an empty sphere (i.e. negative volume)
        # Or (default C_NO_BOUNDING or BOX) Create an empty box (i.e. negative volume)
        bnd = BoundingSphere(Point3d(), -1.0) if boundingType==PoserConst.C_SPHERE_BOUNDING else \
              BoundingBox(Point3d(1.0, 1.0, 1.0), Point3d(-1.0, -1.0, -1.0))

        for faceno in range(0, grp.getNbFace()):
          startIdx = grp.stripCount[faceno]
          lastIdx = grp.stripCount[faceno + 1]
          
          nbEdgeMoved = 0
          faceEdge = [ ]
          faceEdgeIdx = [ ]
          
          for vi in grp.vertIdx[startIdx:lastIdx]: 
            try:
              dp = setFoundDeltas[vi]
              nbEdgeMoved += 1
              # Keep in bounding box, points that have a move
              bnd.combine(grp.coordList[vi])
            except KeyError: #if not found:
              # Record a copy of the point that has no delta              
              faceEdge.append( Point3d(grp.coordList[vi]) )
              faceEdgeIdx.append(vi)
          
          nbmissing = lastIdx - startIdx - nbEdgeMoved
          if (nbEdgeMoved > 0) and (nbmissing > 0):
            # Some vertex are not moved!
            # Get the vertex that compose the "hole face"
            edgePtIdx = grp.vertIdx[startIdx:lastIdx]

            # Calculate '1' loop Vicinity
            edgeLoopPtIdx =  copy.copy(edgePtIdx)

            # First Loop Vicinity
            if useVicinityLoop:
              grp.extendLoopVicinity(edgePtIdx, edgeLoopPtIdx)

            if enhance==PoserConst.C_MLS_ENHANCEMENT:
              # Keep trace of math. error to apply other type of enhancement
              mlsFailed = False

              # Apply deltas to points designed by edgeLoopPtIdx[]
              nuageVoisin = [ ] 
              nuageIdx = [ ] 
              nbPtNuage = 0

              for vi in edgeLoopPtIdx:
                # Add to the cloud (nuage) only vertex that have a delta
                try:
                  faceEdgeIdx.index(vi)
                except ValueError:  #if not found:
                  nuageVoisin.append( Point3d(srcWG.coordList[vi]) )
                  nuageIdx.append(nbPtNuage)
                  try:
                    dp = setFoundDeltas[vi]
                    nuageVoisin[nbPtNuage].add(dp)
                  except KeyError:
                    pass
                  
                  nbPtNuage += 1
              # End for vi

              try:
                # Compute projection Repere with SVD matrix decomposition
                repUVN = Repere(nbPtNuage, nuageIdx, nuageVoisin)

                tabproj = repUVN.project(0, 0, nbPtNuage, nuageIdx, nuageVoisin, None)
                tabdest = repUVN.project(0, 0, nbmissing, nuageIdx, faceEdge, None)

                # Calculate Point Altitudes in the UVN repere
                calcMLS(tabdest, 0, nbmissing, tabproj, alpha)

                ncl = repUVN.reserveProject(0, 0, nbmissing, tabdest, None)

                # Create some new deltas
                for i,vi in enumerate(faceEdgeIdx):
                  if (boundingType==PoserConst.C_NO_BOUNDING) or bnd.intersect(grp.coordList[vi]):
                    dx = round(ncl[i].x - srcWG.coordList[vi].x, 6)
                    dy = round(ncl[i].y - srcWG.coordList[vi].y, 6)
                    dz = round(ncl[i].z - srcWG.coordList[vi].z, 6)
                    if (dx != 0.0) or (dy != 0.0) or (dz != 0.0):
                      dp = DeltaPoint(vi, dx, dy, dz)
                      setNewDeltas[vi]= dp
                    
              except Exception as e:
                logging.warning("Math. Error in %s[%s face=%d] : %s", srcWG.getName(), grp.getName(), faceno, e)
                mlsFailed = True

            if (enhance==PoserConst.C_AVG_ENHANCEMENT) or ((enhance==PoserConst.C_MLS_ENHANCEMENT) and mlsFailed):
              # Find deltas of points designed by edgeLoopPtIdx[]
              dx = 0.0
              dy = 0.0
              dz = 0.0
              nbdelta = 0
              for vi in edgeLoopPtIdx:
                try:
                  dp = setFoundDeltas[vi]
                  dx += dp.x
                  dy += dp.y
                  dz += dp.z
                  nbdelta += 1
                except KeyError:
                  pass

              # Calculage average displacment
              # Value less than 1e-6 are set to 0 by Poser + Null Delta shall not be kept
              if nbdelta>0:
                dx = round(dx / nbdelta, 6)
                dy = round(dy / nbdelta, 6)
                dz = round(dz / nbdelta, 6)
                if (dx != 0.0) or (dy != 0.0) or (dz != 0.0):
                  setNewDeltas.update( { vi:DeltaPoint(vi, dx, dy, dz) for vi in faceEdgeIdx if (boundingType==PoserConst.C_NO_BOUNDING) or bnd.intersect(grp.coordList[vi]) } )

      # Sort the list according to point index (I'm not sure of Poser behavior if not sorted)
      #lstNewDeltas.sort(key=lambda dp: dp.noPt)

    return res



# 
#  * Generic Transformation dedicated to represent : targetGeom <name> valueParm
#  * <name> geomChan <name> xOffsetA|yOffsetA|zOffsetA <name>
#  * xOffset|yOffsetB|zOffsetB <name> taperX|Y|Z <name> scale|X|Y|Z <name>
#  * propagatingScale|X|Y|Z <name> translateX|Y|Z <name> rotateX|Y|Z <name>
#  * smoothScaleX|Y|Z <name> twistX|Y|Z <name> jointX|Y|Z <name> curveX|Y|Z <name>
#  * handGrasp|thumbGrasp|handSpread <name> pointAtParm <name> hairDynamicsParm 
#  * curve  <name>
#  * <name> ... And others ...
#  
#  This method should actually take a ValueOpDelta[] because a single
#  transform can have many valueOpDelta fields!!!
class GenericTransform(StructuredAttribut):
  # Managed operators
  OPS = (PoserToken.E_valueOpDeltaAdd, PoserToken.E_valueOpPlus, PoserToken.E_valueOpMinus, PoserToken.E_valueOpTimes, PoserToken.E_valueOpDivideBy, PoserToken.E_valueOpDivideInto, PoserToken.E_valueOpKey)

  def __init__(self, poserType=None, channelName=None, vodFigure=None, vodActor=None, vodChannel=None):
    super(GenericTransform, self).__init__()

    #  PoserMeshedObject that contains the channel list 
    self._pmo = None

    #  Optimized attributs
    self._transfName = None
    self._initValue = 0.0
    self._hidden = False
    self._enabled = True

    #  V9 attribut (unknown effect)
    self._forceLimits = 4.0
    self._min = -100000
    self._max = 100000
    self._trackingScale = 0.001
    self._interpStyleLocked = -sys.maxsize
    self._keys = None
    # self._indexes = -sys.maxsize - Replaced by a computed value
    self._numbDeltas = -sys.maxsize
    self._dlt = None
    self._lstVOD = [ ]

    self._hasStaticValue = False
    self._staticValue = 0.0

    if poserType:
      self.setPoserType(poserType)
      self.setName(channelName)

      #  20080821: Do not add a prefix to printable name (old P4 Stuff)
      self._transfName = channelName[3:] if (channelName.startswith("PBM") or channelName.startswith("FBM")) else channelName
      self.addKeyFrame(0, 0.0)
      #  Default keys are static=0
      self._interpStyleLocked = 0

      #  Create the deltas table
      if poserType != PoserToken.E_valueParm:
        self._numbDeltas = 0
        #self._dlt = Deltas()

        if vodActor or vodChannel:
          if vodFigure:
            #  HACK HACK HACK Hardwiring valueOpDeltaAdd for now. This will let the
            #  typical case of
            #  valueOpDeltaAdd work for now but all others (valueOpDeltaPlus,
            #  valueOpDeltaMinus, etc)
            #  will be broken! I'll fix this after Olivier adds his hair stuff.
            vod = ValueOpDelta(PoserToken.E_valueOpDeltaAdd, vodFigure, vodActor, vodChannel, 1.0)
            self.addVOD(vod)



  # Overload setPoserType to create Deltas
  def setPoserType(self, poserType):
    StructuredAttribut.setPoserType(self, poserType)
    if not self._dlt and (poserType==PoserToken.E_targetGeom):
      self._dlt = Deltas()

  #
  # Copy all data from src, except the _pmo if parentPO is set
  #
  def copy(self, src, parentPO=None):
    self.setPoserType(self.getPoserType())
    self.setName(src.getName())

    self._pmo = parentPO if parentPO else src._pmo
    self._transfName = src._transfName
    self._initValue = src._initValue
    self._hidden = src._hidden
    self._enabled = src._enabled
    self._forceLimits = src._forceLimits
    self._min = src._min
    self._max = src._max
    self._trackingScale = src._trackingScale
    self._interpStyleLocked = src._interpStyleLocked
    self._hasStaticValue = src._hasStaticValue
    self._staticValue = src._staticValue
    self._numbDeltas = src._numbDeltas
    self._keys = Keys(src=src._keys)
    self._dlt = Deltas(src=src._dlt)
    self._lstVOD = [ ValueOpDelta(src=v) for v in src.getVOD() ]
    self._transfName = src._transfName


  @classmethod
  def createFrom(cls, gtdef:'list from PoserToken'):
    ''' Create a GenericTransform and initialize it with the data
    of tables PoserToken.MINIMAL_***_CHANNEL
    
    Package Reserved method
    '''
    vp = GenericTransform(gtdef[0], gtdef[0].defaultName)
    
    vp.setPrintName(gtdef[0].token)
    vp.setInterpStyleLocked(0)
    vp.setMin(gtdef[2])
    vp.setMax(gtdef[3])
    vp.setInitValue(gtdef[4])      
    vp.setForceLimits(0)
    vp.setTrackingScale(gtdef[8])
    
    if gtdef[1]: #If Static channel
      vp.setStaticValue(gtdef[4])
      vp.setHidden(True)
      
    if (gtdef[0] in PoserToken.SCALES) or not gtdef[1]:
      # Init Scales at 1.0 (100%) and other with their initvalue
      vp.addKeyFrame(0, gtdef[4])
      
    return vp
    

  def addKeyFrame(self, frameno, v):
    if not self._keys:
      self._keys = Keys() #  Default keys are static=0

    self._keys.addKey(frameno, v)


  def getKeysFactor(self, frameNo):
    return self._keys.getKeyFactor(frameNo) if self._keys else 0.0

  def getDeltas(self):
    return self._dlt

  def ishasDeltas(self): return self._dlt and (len(self._dlt.deltaSet)>0)

  # 
  #    * Remove the deltas of tha current channel
  #    
  def removeDeltas(self):
    #  20090711 : Not sure that should be REALLY cleared?
    #  As of 20100226 do not delete anymore the calculation dependencies : _lstVOD.clear();
    #self._indexes = 0
    self._numbDeltas = 0
    #  20190105 : Fix for robusness
    if self.ishasDeltas():
      self.getDeltas().deltaSet.clear()

  # 
  # Optimize deltas : Delete delta with a too small norme
  #    
  def optimizeDeltas(self, refNorm):
    dl = self.getDeltas()
    oldLength = len(dl.deltaSet)
    optSet = { no:t for no,t in dl.deltaSet.items() if t.norme() > refNorm }

    # optTabFinal = Arrays.copyOf(optTab, nbpt)
    dl.deltaSet = optSet
    indexes = len(optSet)
    logging.info("Channel[%s] reduced from %d to %d", self.getName(), oldLength, indexes )
    return indexes 

  def getVOD(self): return self._lstVOD

  # Add an operator to the GenericChannel
  def addVOD(self, vop):
    self._lstVOD.append(vop)

  def findValueOp(self, vodType:'ValueOpDelta', targetName:'str'=None):
    return next((vop for vop in self._lstVOD \
                if (vop.getPoserType()==vodType) and ((targetName is None) or (vop.getChannelName()==targetName))), None)

  # 
  #    * Delete any reference to the given GenericTransform
  #    * @param gt
  #    
  def deleteChannelRef(self, gt):
    chName = gt.getName()
    grpName = gt.getPoserMeshedObject().getName()
    self._lstVOD[:] = [ vod for vod in self._lstVOD if (grpName!=vod.getGroupName()) or (chName!=vod.getChannelName()) ]


  # 
  #    * Delete any reference to the given GenericTransform
  #    * @param gt
  #    
  def deleteFigureRef(self, fig):
    fi = fig.getBodyIndex()
    self._lstVOD[:] = [ vod for vod in self._lstVOD if index(vod.getGroupName())!=fi ]


  # 
  #    * Change the name of referenced part (actor, prop, hairProp, controlProp)  
  #    * @param oldPartName
  #    * @param newPartName
  #    
  def changeReference(self, oldPartName, newPartName, changeRefToBODY=True):
    for vod in self._lstVOD:
      if not vod.getGroupName().startswith(PoserConst.C_BODY) or changeRefToBODY:
        vod.setGroupName(TAS(vod.getGroupName(), oldPartName, newPartName))
      
    for po in self.getLstAttr():
      if po.getPoserType() == PoserToken.E_otherActor:
        po.TAS(oldPartName, newPartName)


  def addAttribut(self, po):
    if po.getPoserType() in GenericTransform.OPS:
      self._lstVOD.append(po)
    elif po.getPoserType()==PoserToken.E_keys:
      self._keys = po
    elif po.getPoserType()==PoserToken.E_deltas:
      self._dlt = po
    else:
      super(GenericTransform, self).addAttribut(po)

  def setDirect(self, tokenID, val):
    if tokenID==PoserToken.E_name:
      self._transfName = val
    elif tokenID==PoserToken.E_initValue:
      self._initValue = float(val)
    elif tokenID==PoserToken.E_hidden:
      self._hidden = val.startswith("1")
    elif tokenID==PoserToken.E_trackingScale:
      self._trackingScale = float(val)
    elif tokenID==PoserToken.E_min:
      self._min = float(val)
    elif tokenID==PoserToken.E_max:
      self._max = float(val)
    elif tokenID==PoserToken.E_forceLimits:
      self._forceLimits = float(val)
    elif tokenID==PoserToken.E_flipped:
      #  Bad fix ... but for the moment
      self.addAttribut(OffSA("flipped"))
    elif tokenID==PoserToken.E_indexes:
      #self._indexes = int(val)
      pass
    elif tokenID==PoserToken.E_numbDeltas:
      self._numbDeltas = int(val)
    elif tokenID==PoserToken.E_enabled:
      self._enabled = val.startswith("1")
    elif tokenID==PoserToken.E_interpStyleLocked:
      self._interpStyleLocked = int(val)
    elif tokenID==PoserToken.E_staticValue:
      self.setStaticValue(float(val))
    else:
      logging.warning("Unexpected Direct in GenericTranform:%s (%s)", tokenID.token, val)

  #    * @see pftk.poser.kern.StructuredAttribut#write(java.io.PrintWriter,
  def write(self, fw, pfx):
    fw.write(f"{pfx}{self.getPoserType().token} {self.getName()}\n")
    nextPfx = pfx + "\t"
    fw.write(f"{nextPfx}{{\n")
    fw.write(f"{nextPfx}name {self._transfName}\n")
    fw.write(f"{nextPfx}initValue {self._initValue}\n")
    WBoolLine(fw, nextPfx, "hidden", self._hidden )
    
    #  Test Poser language version before writing
    if self._pmo and (self._pmo.getPoserFile().isVersion(PoserConst.POSER_V9)):
      WBoolLine(fw, nextPfx, "enabled", self._enabled )
    fw.write(f"{nextPfx}forceLimits {self._forceLimits}\n")
    fw.write(f"{nextPfx}min {self._min}\n")
    fw.write(f"{nextPfx}max {self._max}\n")
    fw.write(f"{nextPfx}trackingScale {self._trackingScale}\n")
    #  Print Keys
    if self._keys != None:
      self._keys.write(fw, nextPfx)
    if self._interpStyleLocked != -sys.maxsize:
      fw.write(f"{nextPfx}interpStyleLocked {self._interpStyleLocked}\n")
    if self._hasStaticValue:
      fw.write(f"{nextPfx}staticValue {self._staticValue}\n")
    for vod in self._lstVOD:
      vod.write(fw, nextPfx)
    for po in self._lstAttr:
      po.write(fw, nextPfx)
    if self._dlt and (len(self._dlt.deltaSet)>0):
      fw.write(f"{nextPfx}indexes {len(self._dlt.deltaSet)}\n")
      fw.write(f"{nextPfx}numbDeltas {self._numbDeltas}\n")
      self._dlt.write(fw, nextPfx)
    fw.write(f"{nextPfx}}}\n")



  def setPoserMeshedObject(self, pmo):
    self._pmo = pmo

  def getPoserMeshedObject(self): return self._pmo

  def getPrintName(self): return self._transfName

  def getQualifiedName(self):
    return (self._pmo.getName() if self._pmo else '?' )+ '.' + self.getName()

  def setPrintName(self, n):
    self._transfName = n

  def isHidden(self): return self._hidden

  def setHidden(self, h):
    self._hidden = h

  def getMin(self): return self._min

  def getMax(self): return self._max

  def setMin(self, m):
    self._min = m

  def setMax(self, m):
    self._max = m

  def setInterpStyleLocked(self, isl):
    self._interpStyleLocked = isl

  def getPrintableDependencies(self):
    return [ vod.getPoserType().token + " " +vod.getGroupName() + " " + vod.getChannelName() for vod in self.getVOD() ]

  def getTrackingScale(self): return self._trackingScale

  def setTrackingScale(self, m):
    self._trackingScale = m

  def setNumbDeltas(self, n):
    self._numbDeltas = n

  def getNumbDeltas(self): return self._numbDeltas

  def getInitValue(self): return self._initValue

  def setInitValue(self, m):
    self._initValue = m

  def getStaticValue(self): return self._staticValue

  def setStaticValue(self, m):
    self._hasStaticValue = True
    self._staticValue = m

  def getForceLimits(self): return self._forceLimits

  def setForceLimits(self, m):
    self._forceLimits = m

  def setDeltaTab(self, tdp):
    self._dlt.deltaSet = tdp
    

  def getKeys(self): return self._keys

  # Convert a GenericTransform to a static one (i.e. no keys)
  def toStatic(self):
    ret = C_OK
    if self._hasStaticValue:
      logging.info("Channel[%s]: keep static=%d", self.getName(), self.getStaticValue())
    else:
      self.setStaticValue(self.getInitValue())
      logging.info("Channel[%s]: set static=%d", self.getName(), self.getStaticValue())

    if self._keys:
      ret = self._keys.toStatic(self.getStaticValue())
    return ret

  # Determine if a target morph shall be taken into account. 
  # @param attrName        Name of the targetGeom or valueParm
  # @param setTargetMorph  Set of names
  # @return    true when the set of target morph is empty AND the attrName starts with "PBM" OR when the set of target contains the given attrName
  @classmethod
  def concerned(cls, attrName, setTargetMorph):
    return attrName.startswith("PBM") if setTargetMorph==None else attrName in setTargetMorph 


class Scale(GenericTransform):

  def __init__(self):
    super(Scale, self).__init__()
    self.ambiguious = True
    self._isSimple = True

  #    * @see deyme.v3d.poser.PoserObject#write(java.io.FileWriter)
  #    
  def write(self, fw, pfx):
    if self._isSimple:
      fw.write(pfx + self.getPoserType().token + " " + self.getName() + '\n')
    else:
      super(Scale, self).write(fw, pfx)

  def addAttribut(self, po):
    super(Scale, self).addAttribut(po)
    self._isSimple = False

# 
# DepthMapSize attribut.
#  
class DepthMapSize(GenericTransform):
  def __init__(self):
    super(DepthMapSize, self).__init__()
    self.ambiguious = True
    self._isSimple = True

  #    * @see deyme.v3d.poser.PoserObject#write(java.io.FileWriter)
  def write(self, fw, pfx):
    if self._isSimple:
      fw.write(f"{pfx}{self.getPoserType().token} {self.getName()}\n")
    else:
      super(DepthMapSize, self).write(fw, pfx)

  def addAttribut(self, po):
    super(DepthMapSize, self).addAttribut(po)
    self._isSimple = False

# 
# Represents the list of channels of a PoserMeshedObject.
# class Channels(StructuredAttribut):  --> Inlined in PoserMeshedObject

# 
#  * Used to create morph. 
#  
class ChannelMorphStatus(object):
  #    
  PRESENCE_IMG = ("Unused", "in Reference", "in Destination", "Both")
  ATTR = ("Channel", "Presence", "RefFactor", "RefNonNullDesc", "RefHasDelta", "SrcFactor", "SrcNonNullDesc", "SrcHasDelta", "Optimized")

  def __init__(self, cnOrgt, InitSetTargetMorph=None):

    # 
    # * 0x0 = Unused
    # * 0x1 = Used only in Reference figure
    # * 0x2 = Used only in Destination figure
    # * 0x3 = Used by both
    #    
    self.presence = 0
    self.peer = None
    self.hasDelta = False
    
    if isinstance(cnOrgt, GenericTransform):
      self.channelName = cnOrgt.getName()
      self.selected = GenericTransform.concerned(cnOrgt.getName(), InitSetTargetMorph)
      self.finalFactor = cnOrgt.getKeysFactor(0)
      self.updateDelta(cnOrgt)

    else: # cnOrgt is a String
      self.channelName = cnOrgt
      self.finalFactor = 0.0
      self.hasNonNullDesc = False     
      self.selected = False


  def updateDelta(self, attrDesc):
    dlt = attrDesc.getDeltas()
    self.hasDelta = self.hasDelta or dlt

  def getChannelName(self): return self.channelName

  def setChannelName(self, channelName):
    self.channelName = channelName

  def getFinalFactor(self): return self.finalFactor

  def setFinalFactor(self, finalFactor):
    self.finalFactor = finalFactor

  def isHasNonNullDesc(self): return self.hasNonNullDesc

  def setHasNonNullDesc(self, hasNonNullDesc):
    self.hasNonNullDesc = hasNonNullDesc

  def isHasDelta(self): return self.hasDelta

  def setHasDelta(self, hasDelta):
    self.hasDelta = hasDelta

  def isSelected(self): return self.selected

  def setSelected(self, selected):
    self.selected = selected

  def getPeer(self): return self.peer

  def setPeer(self, peer):
    self.peer = peer

  def getPresence(self): return self.presence

  def setPresence(self, presence):
    self.presence = presence


# 
# A list of ChannelMorphStatus
#  
class ChannelMorphStatusList(list):
  FILTER_NONE = 0x0000
  FILTER_NONNULLREFDELTA = 0x0001
  FILTER_REFPRESENCE = 0x0002
  FILTER_NONNULLREFFACTOR = 0x0004


  # 
  # Create an empty list.
  # Create an filtered copy of list lstsrc.
  #    * @param filter
  #    
  # def __init___0(self, filterCode)
  # def __init__(self):
  def __init__(self, filterCode=None, lstsrc=None):
    super(ChannelMorphStatusList, self).__init__()
    self._filter = filterCode if filterCode else ChannelMorphStatusList.FILTER_NONE
    if lstsrc:
      for refCS in lstsrc:
        self.addFiltered(refCS, None)

  def __str__(self):
    return str(len(self))

  def find(self, channelName):
    for c in self:
      if channelName == c.channelName:
        return c
    return None

  def extract(self, channelName):
    for c in self:
      if channelName == c.channelName:
        self.remove(c)
        return c

    return None

  def addFiltered(self, refCS, srcCS):
    refCS.peer = srcCS
    result = True
    if self._filter != self.FILTER_NONE:
      if (self._filter & self.FILTER_NONNULLREFDELTA) == self.FILTER_NONNULLREFDELTA:
        result = result and refCS.hasDelta
      if (self._filter & self.FILTER_REFPRESENCE) == self.FILTER_REFPRESENCE:
        result = result and ((refCS.presence & 0x1) == 0x1)
      if (self._filter & self.FILTER_NONNULLREFFACTOR) == self.FILTER_NONNULLREFFACTOR:
        result = result and (refCS.finalFactor != 0.0)
    if result:
      self.append(refCS)

  def getFilter(self): return self._filter

  # 
  # Return a Set with the selected channels.
  #    
  def getChannelSet(self):
    return { cs.channelName for cs in self if cs.selected }

  # 
  #  Save the list of selected channels.
  #  @param fn  Full path name of the file to create.
  #TODO: Useless and serialization pb due to the link to a Generictransform
#   def saveSelected(self, fn):
#     try:
#       #  Write File Version
#       ois = open(fn, "w")
# 
#       jsonD = { cs.channelName : cs.__dict__ for cs in self if cs.selected }
# 
#       json.dump(jsonD, ois)
# 
#       ois.close()
#       logging.info("List saved in %s (JSON)", fn)
#     except IOError as ioex:
#       ioex.printStackTrace()
# 
#   @classmethod
#   def loadList(cls, fn):
#     l = ChannelMorphStatusList()
#     try:
#       ois = open(fn, "r")
#       data = json.load(ois)
#       for cs in data.values():
#         l.append(cs)
# 
#       ois.close()
#     except IOError as ioex:
#       # ioex.printStackTrace()
#       print(str(ioex))
#       l = None
#     return l


# 
# This class represents the "frontImage" & "sideImage" elements of Poser files.
#  
class FrontImage(StructuredAttribut):
  def __init__(self):
    super(FrontImage, self).__init__()
    self._loaded = False
    self._file = None #  "NO_MAP" 
    self._flipped = False

  def setDirect(self, tokenID, val):
    if tokenID==PoserToken.E_loaded:
      self._loaded = val.startswith("1")
    elif tokenID==PoserToken.E_file:
      self._file = None if val == PoserConst.C_NO_MAPG or val == PoserConst.C_NO_MAP else val
    elif tokenID==PoserToken.E_flipped:
      self._flipped = val.startswith("1")
    else:
      logging.warning("Unexpected Direct in FrontImage:%s (%s)", tokenID.token, val)

  #    * @see deyme.v3d.poser.PoserObject#write(java.io.FileWriter)
  def write(self, fw, pfx):
    fw.write(f"{pfx}{self.getPoserType().token} {self.getName()}\n")
    nextPfx = pfx + "\t"
    fw.write(f"{nextPfx}{{\n")
    fw.write(f"{nextPfx}loaded {1 if self._loaded else 0}\n")
    fw.write(f"{nextPfx}file {self._file if self._file else PoserConst.C_NO_MAPG}\n")
    fw.write(f"{nextPfx}flipped {1 if self._flipped else 0}\n")
    for po in self._lstAttr:
      po.write(fw, nextPfx)
    if not self._lstAttr:
      fw.write('\n')
    fw.write(f"{nextPfx}}}\n")

# package: pftk.poser.kern
class SideImage(FrontImage):
  def __init__(self):
    super(SideImage, self).__init__()



class WeightMap(StructuredAttribut):
  def __init__(self, Name=None):
    super(WeightMap, self).__init__(Name)
    self._numbVerts = 0
    self._vertTab = []

  # Overloaded reader    
  def read(self, st):
    self._vertTab = [ ]

    while True:
      code,cn,rw = st.getLine()
      if code== PoserFileParser.TT_WORD:
        if cn=="v":
          tv = rw.split()
          self._vertTab.append( ( int(tv[0]), float(tv[1]) ) )
        elif cn=='numbVerts':
          self._numbVerts = int(rw)
        else: #  Mot illegal
          logging.info("Line[%s] - Not accepted word:%s",st.lineno(), cn)
      elif (code==PoserFileParser.TT_RIGHTBRACKET) or (code==PoserFileParser.TT_EOF): 
        break
      else:
        logging.warning("Line[%s] - Not Accepted :%s", st.lineno(), cn)
        raise ParsingErrorException()

    # End of read

  def write(self, fw, pfx):
    fw.write(f"{pfx}{self.getPoserType().token} {self.getName()}\n")
    nextPfx = pfx + "\t"
    fw.write(f"{nextPfx}{{\n")
    fw.write(f"{nextPfx}numbVerts {self._numbVerts}\n")
    for vi in self._vertTab:
      fw.write(f"{pfx}\tv {vi[0]:d} {vi[1]:9.6f}\n")

    fw.write(f"{pfx}\t}}\n")
 


class Zones(StructuredAttribut):
  def __init__(self, Name=None):
    super(Zones, self).__init__(Name)
   
class WeightMapZone(StructuredAttribut):
  def __init__(self, Name=None):
    super(WeightMapZone, self).__init__()
   
class SphereZone(StructuredAttribut):
  def __init__(self, Name=None):
    super(SphereZone, self).__init__()

class CapsuleZone(StructuredAttribut):
  def __init__(self, Name=None):
    super(CapsuleZone, self).__init__(Name)


# E_Data Tuple Key/Value for Customer Data Storage
custData = namedtuple('custData', ['val', 'withPose', 'withMat'])

class CustomData(StructuredAttribut):
  ''' Represents the customData structure that can be found at figure/actor/prop level '''
  def __init__(self, Name=None):
    super(CustomData, self).__init__(Name)
    self._data = {}

  def set(self, key, val, withPose=0, withMat=0):
    self._data[key] = custData(val, withPose, withMat)

  def get(self, key):
    return self._data.get(key, None)

  # Overloaded reader    
  def read(self, st):
    while True:
      code,cn,rw = st.getLine()
      if code== PoserFileParser.TT_WORD:
        if cn=="data":
          tv = rw.split()
          self.set(tv[0], RemoveQuotes(tv[3]), int(tv[1]), int(tv[2]))
        else: #  Mot illegal
          logging.info("Line[%s] - Not accepted word:%s",st.lineno(), cn)
      elif (code==PoserFileParser.TT_RIGHTBRACKET) or (code==PoserFileParser.TT_EOF): 
        break
      else:
        logging.warning("Line[%s] - Not Accepted :%s", st.lineno(), cn)
        raise ParsingErrorException()
    # End of read

  def write(self, fw, pfx):
    fw.write(f"{pfx}{self.getPoserType().token} {self.getName()}\n")
    nextPfx = pfx + "\t"
    fw.write(f"{nextPfx}{{\n")

    nextPfx = pfx + '\tdata {0:s} {1:d} {2:d} "{3:s}"\n'

    for k,d in self._data.items():
      fw.write(nextPfx.format(k, d.withPose, d.withMat, d.val))
 
    fw.write(f"{pfx}\t}}\n")


# 
# Calculate point mapping table :<br>
# For each vertex of srcWG, find the closest vertex in the list of reference WaveGeom 
# If maxDist is not null, some vertex could have <b>no</b> mapping. 
# Should this occur, refGeom attribut of PtMapping is set to null. 
# 
# @param srcWG           Source WaveGeom
# @param lstRefGeom      List of reference WaveGeom
# @param translation     Translation between original MeshedObject and its localisation in this.
# @param maxDist         Maximal distance to take a point into account. If null, no maxDist are applied.
# @return                A Table of non null PtMapping
def calcMapping_KDTree(srcWG, lstRefGeom, translation, maxDist):

    vt = translation if translation else Vector3d()
    tabGeom = [ ]
    tabStart = [ ]

    # Create a numpy table Nx3
    nbsrc=0
    for refGC in lstRefGeom:
      refWG = refGC.getWaveGeom()
      nbsrc += len(refWG.coordList)


    refNo = 0
    npTab = np.zeros( (nbsrc, 3) )
    for refGC in lstRefGeom:
      refWG = refGC.getWaveGeom()
      startNo = refNo
      for p in refWG.coordList:
        npTab[refNo] = [ p.x, p.y, p.z ]
        refNo+=1
        tabGeom.append(refGC)
        tabStart.append(startNo)

    # Create an KDtree with the numpy table
    tree = spatial.KDTree(npTab, leafsize=10 if nbsrc<10000 else 100)

    svect = np.zeros((len(srcWG.coordList),3))
    for i,psrc in enumerate(srcWG.coordList):
      svect[i] = [ psrc.x - vt.x , psrc.y - vt.y , psrc.z - vt.z ]

    rest, resIdx = tree.query( svect, distance_upper_bound=maxDist )

    tabMapping = [ PtMapping(i, int(ri-tabStart[ri]), tabGeom[ri], float(rest[i])) \
                  for i,ri in enumerate(resIdx) if rest[i]<maxDist ]
    
    return tabMapping

# 
# Filter the result list to remove too short vectors.
# Does not modify the source set
#    
def filterLength(setDeltas, minVectLen):
  return { dp.noPt:dp for dp in setDeltas.values() if minVectLen <= dp.norme() }

# 
# @param tabMapping
# @param refFigure
# @param channelName
# @param minVectLen
# @return a set of { noPt:DeltaPoint }
#    
# findNewDelta(PtMapping[] tabMapping, Figure refFigure, String channelName)
# findNewDelta(PtMapping[] tabMapping, PoserMeshedObject refMeshedObj, String channelName)
def findNewDelta(tabMapping, refObj, channelName):
    setNewDeltas = { }
    emptySet = { }

    #if isinstance(refObj, "Figure"):
    if refObj.__class__.__name__ == "Figure":
      lastGeom = None
      for pm in tabMapping:
        if pm.refGeom:
          if pm.refGeom!=lastGeom:
            lastGeom = pm.refGeom
            gt = refObj.getPoserFile().findActor(pm.refGeom.getName()).getTargetGeom(channelName)            
            cachDelta = gt.getDeltas().deltaSet if gt else emptySet
              
          #srcDp = refObj.getPoserFile().findActor(pm.refGeom.getName()).findDelta(channelName, pm.refNo)
          try:
            srcDp = cachDelta[pm.refNo]
            dp = DeltaPoint(pm.srcNo, srcDp.x, srcDp.y, srcDp.z)
            setNewDeltas[dp.noPt] = dp
          except KeyError:
            pass
    else: # refObj is supposed to be a PoserMeshedObject
      gt = refObj.getTargetGeom(channelName)

      # FIX: More robust
      if gt:
        cachDelta = gt.getDeltas().deltaSet if gt.getDeltas() else emptySet
  
        for pm in tabMapping:
          if pm.refGeom:
            try:
              srcDp = cachDelta[pm.refNo]
              dp = DeltaPoint(pm.srcNo, srcDp.x, srcDp.y, srcDp.z)
              setNewDeltas[dp.noPt] = dp
            except KeyError:
              pass

    return setNewDeltas


# Increase recursion limit for very big data sets - For KDTree
sys.setrecursionlimit(20000)
